#pragma once

#include "RTTI.h"
#include "vec4.h"
#include "mat4.h"

#include <string>
#include <sstream>
#include <stdio.h>
#include "RTTI.h"


//Forward Declarations
class Scope;

class Datum
{
	public:
		typedef Datum *ReferenceType;
		typedef void *PointerType;
		typedef Scope *ScopeType;
		typedef RTTI *RTTIType;
		enum TypeE
		{
			TYPE_UNKNOWN,
			TYPE_INT,
			TYPE_FLOAT,
			TYPE_VECTOR,
			TYPE_MATRIX,
			TYPE_REFERENCE,
			TYPE_SCOPE,
			TYPE_STRING,
			TYPE_POINTER,
			TYPE_RTTI
		};

		Datum();
		Datum(const Datum& source);
		Datum& operator=(const Datum& source);
		~Datum();
		void SetType(const TypeE newType);
		const TypeE GetType() const;
		void SetNumValues(unsigned int numberOfValues);
		const unsigned int Size() const;
		const unsigned int Capacity() const;
		void SetStorage(void* array, unsigned int numberOfValues, unsigned int capacity);
		bool operator==(const Datum & otherDatum) const;
		bool operator!=(const Datum & otherDatum) const;
		void Clear();

		// Public value accessors - Provide one signature for each type, one for each primitive type. 
		
		//These are convenience routines for when Datum is a scalar.
		// int
		bool operator==( const int   & that ) const;
		bool operator!=( const int   & that ) const;
		// float 
		bool operator==( const float & that ) const;
		bool operator!=( const float & that ) const;
		// vector
		bool operator==( const Vec4 & that ) const;
		bool operator!=( const Vec4 & that ) const;
		// matrix
		bool operator==( const Mat4 & that ) const;
		bool operator!=( const Mat4 & that ) const;
		// reference
		bool operator==( const ReferenceType & that ) const;
		bool operator!=( const ReferenceType & that ) const;
		// Scope
		bool operator==( const ScopeType & that ) const;
		bool operator!=( const ScopeType & that ) const;
		// string
		bool operator==( const std::string & that ) const;
		bool operator!=( const std::string & that ) const;
		// pointer
		bool operator==( const PointerType that ) const;
		bool operator!=( const PointerType that ) const;

		/*************************************
			Set
		**************************************/
		//Provide one signature for each type

		// int
		void Set(int &data, unsigned int index = 0 );
		void Set(const int &data, unsigned int index = 0 );
		// float
		void Set(float &data, unsigned int index = 0 );
		void Set(const float &data, unsigned int index = 0 );
		// vector
		void Set(Vec4 &data, unsigned int index = 0 );
		void Set(const Vec4 &data, unsigned int index = 0 );
		// matrix
		void Set(Mat4 &data, unsigned int index = 0 );
		void Set(const Mat4 &data, unsigned int index = 0 );
		// reference
		void Set(ReferenceType &data, unsigned int index = 0 );
		void Set(const ReferenceType &data, unsigned int index = 0 );
		// Scope
		void Set(ScopeType &data, unsigned int index = 0 );
		void Set(const ScopeType &data, unsigned int index = 0 );
		// string
		void Set(std::string &data, unsigned int index = 0 );
		void Set(const std::string &data, unsigned int index = 0 );
		// pointer
		void Set(const PointerType data );
		// rtti
		void Set(RTTIType data,  unsigned int index = 0);

		// Assignment operator - Wrappers for Set that make Datum objects appear syntactically like scalar variables with dynamic type.  
		// int
		Datum & operator=( const int  & iVal );
		// float
		Datum & operator=( const float & fVal ); 
		// vector
		Datum & operator=( const Vec4 & vVal ); 
		// matrix
		Datum & operator=( const Mat4 & vVal ); 
		// reference
		Datum & operator=( const ReferenceType & rVal );
		// Scope
		Datum & operator=( const ScopeType & sVal );
		// string
		Datum & operator=( const std::string & sVal );
		// pointer
		Datum & operator=( const PointerType pVal ); 
		
		/*************************************
			Gets
		**************************************/
		//Provide one signature for each type

		template <typename T> const T& Get( const unsigned int index = 0 ) const { return(0); } // Unknown
		// int
		template <> const int& Get<int>( const unsigned int index) const { 
			assert(mType == TYPE_INT); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_INT[index]);
		}
		// float
		template <> const float& Get<float>( const unsigned int index ) const { 
			assert(mType == TYPE_FLOAT); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_FLOAT[index]);
		} 
		// vector
		template <> const Vec4& Get<Vec4>( const unsigned int index ) const { 
			assert(mType == TYPE_VECTOR); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_VECTOR[index]);
		} 
		// matrix
		template <> const Mat4& Get<Mat4>( const unsigned int index ) const { 
			assert(mType == TYPE_MATRIX); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_MATRIX[index]);
		} 
		// reference
		template <> const ReferenceType& Get<ReferenceType>( const unsigned int index ) const { 
			assert(mType == TYPE_REFERENCE); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_REFERENCE[index]);
		}
		// Scope
		template <> const ScopeType& Get<ScopeType>( const unsigned int index ) const { 
			assert(mType == TYPE_SCOPE); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_SCOPE[index]);
		}
		// string
		template <> const std::string& Get<std::string>( const unsigned int index ) const { 
			assert(mType == TYPE_STRING); // In debug builds, verify that the given value type matches the type of this object
			assert(index < mSize);
			return(mData.TYPE_STRING[index]);
		} 
		// pointer
		template <> const PointerType& Get<PointerType>( const unsigned int index ) const { 
			assert(mType == TYPE_POINTER); // In debug builds, verify that the given value type matches the type of this object
			return(mData.TYPE_POINTER);
		} 
		// rtti
		template <> const RTTIType& Get<RTTIType>( const unsigned int index ) const { 
			assert(mType == TYPE_RTTI); // In debug builds, verify that the given value type matches the type of this object
			return(mData.TYPE_RTTI[index]);
		}
	

		/*************************************
			GetAs
		**************************************/
		
		template <typename T> const T GetAs( const unsigned int index = 0 ) const {} // Unknown

		//static int GetAsInt( std::string value );

		//! GetAs<int> 
		/*!
			Gets the data in the datum as if it were an int
			\param - the index of the data to convert and return
			\return - the data as an int
		*/
		template <> const int GetAs<int>( const unsigned int index) const { 
			assert(index < mSize);
			assert((mType == TYPE_INT) || (mType == TYPE_FLOAT) || (mType == TYPE_STRING));	

			switch(mType)
			{
				case TYPE_INT: return (mData.TYPE_INT[index]);
				case TYPE_FLOAT: return (mData.TYPE_FLOAT[index]);
				case TYPE_STRING:
					{
						//return GetAsInt(mData.TYPE_STRING[index]);
						assert(mData.TYPE_STRING[index].empty() == false);
						return (atoi(mData.TYPE_STRING[index].c_str()));
					}
					break;
				default: return(mData.TYPE_INT[index]);	
			}
		}
		
		//! GetAs<float> 
		/*!
			Gets the data in the datum as if it were an float
			\param - the index of the data to convert and return
			\return - the data as an float
		*/
		template <> const float GetAs<float>( const unsigned int index) const { 
			assert(index < mSize);
			assert((mType == TYPE_INT) || (mType == TYPE_FLOAT) || (mType == TYPE_STRING));	

			switch(mType)
			{
				case TYPE_INT: return (mData.TYPE_INT[index]);
				case TYPE_FLOAT: return (mData.TYPE_FLOAT[index]);
				case TYPE_STRING:
					{
						assert(mData.TYPE_STRING[index].empty() == false);
						return (atof(mData.TYPE_STRING[index].c_str()));
					}
					break;
				default: return(mData.TYPE_FLOAT[index]);	
			}
		}
		
		//! GetAs<string> 
		/*!
			Gets the data in the datum as if it were an string
			\param - the index of the data to convert and return
			\return - the data as an string
		*/
		template <> const std::string GetAs<std::string>( const unsigned int index) const { 
			assert(index < mSize);
			assert((mType == TYPE_INT) || (mType == TYPE_FLOAT) || (mType == TYPE_STRING) || (mType == TYPE_VECTOR) || (mType == TYPE_MATRIX));	

			std::string temp;
			char myStringBuffer[60];

			switch(mType)
			{
				case TYPE_INT:
					{ 
						int tempInt = mData.TYPE_INT[index];
						sprintf(myStringBuffer, "%d", tempInt);
						temp.append(myStringBuffer);
						return(temp);
					}
					break;
				case TYPE_FLOAT:
					{ 
						float tempFloat = mData.TYPE_FLOAT[index];
						sprintf(myStringBuffer, "%lf\0", tempFloat);
						temp.append(myStringBuffer);
						return temp;
					}
					break;
				case TYPE_VECTOR:
					{ 
						Vec4 tempVector = mData.TYPE_VECTOR[index];
						sprintf(myStringBuffer, "%f %f %f %f", tempVector.x, tempVector.y, tempVector.z, tempVector.w);
						temp.append(myStringBuffer);
						return(temp);
					}
					break;
				case TYPE_MATRIX:
					{ 
						Mat4 tempMatrix = mData.TYPE_MATRIX[index];
						sprintf(myStringBuffer, "%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", 
						tempMatrix.m[0][0], tempMatrix.m[0][1], tempMatrix.m[0][2], tempMatrix.m[0][3],
						tempMatrix.m[1][0], tempMatrix.m[1][1], tempMatrix.m[1][2], tempMatrix.m[1][3],
						tempMatrix.m[2][0], tempMatrix.m[2][1], tempMatrix.m[2][2], tempMatrix.m[2][3],
						tempMatrix.m[3][0], tempMatrix.m[3][1], tempMatrix.m[3][2], tempMatrix.m[3][3]);
						temp.append(myStringBuffer);
						return(temp);
					}
					break;
				default: return(mData.TYPE_STRING[index]);	
			}
		}
		
		//! GetAs<Mat4> 
		/*!
			Gets the data in the datum as if it were an int
			\param - the index of the data to convert and return
			\return - the data as an int
		*/
		template <> const Mat4 GetAs<Mat4>( const unsigned int index) const { 
			assert(index < mSize);
			assert(mType == TYPE_STRING);	

			char * pEnd;
			//char numbers[] = mData.TYPE_STRING[index].c_str();

			switch(mType)
			{
				case TYPE_STRING:
					{
						assert(mData.TYPE_STRING[index].empty() == false);
						
						Mat4 newMatrix;
						float value = strtod(mData.TYPE_STRING[index].c_str(), &pEnd);
						newMatrix.m[0][0] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[0][1] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[0][2] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[0][3] = value;
						
						value = strtod(pEnd, &pEnd);
						newMatrix.m[1][0] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[1][1] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[1][2] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[1][3] = value;

						value = strtod(pEnd, &pEnd);
						newMatrix.m[2][0] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[2][1] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[2][2] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[2][3] = value;

						value = strtod(pEnd, &pEnd);
						newMatrix.m[3][0] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[3][1] = value;
						value = strtod(pEnd, &pEnd);
						newMatrix.m[3][2] = value;
						value = strtod(pEnd, NULL);
						newMatrix.m[3][3] = value;

						return (newMatrix);
					}
					break;
				default: return(mData.TYPE_MATRIX[index]);	
			}
		}
		
		//! GetAs<vec> 
		/*!
			Gets the data in the datum as if it were an float
			\param - the index of the data to convert and return
			\return - the data as an float
		*/
		template <> const Vec4 GetAs<Vec4>( const unsigned int index) const { 
			assert(index < mSize);
			assert(mType == TYPE_STRING);	

			char * pEnd;
			//char numbers[] = mData.TYPE_STRING[index].c_str();

			switch(mType)
			{
				case TYPE_STRING:
					{
						assert(mData.TYPE_STRING[index].empty() == false);
						
						Vec4 newVector4;
						float value = strtod(mData.TYPE_STRING[index].c_str(), &pEnd);
						newVector4.x = value;
						value = strtod(pEnd, &pEnd);
						newVector4.y = value;
						value = strtod(pEnd, &pEnd);
						newVector4.z = value;
						value = strtod(pEnd, NULL);
						newVector4.w = value;

						return (newVector4);
					}
					break;
				default: return(mData.TYPE_VECTOR[index]);	
			}
		}




		void RemoveIndex(const unsigned int index);

		/*************************************
			Static utility routines
		**************************************/
		static std::string StringFromType(TypeE enumType);
		static TypeE TypeFromString(std::string typeName);
		static bool HasValue(Datum& myDatum, unsigned int index = 0);
		static bool ReferenceHasValue(Datum& myDatum, unsigned int index = 0);
		
		union DataType
		{
			void* TYPE_UNKNOWN;
			int* TYPE_INT;
			float* TYPE_FLOAT;
			Vec4* TYPE_VECTOR;
			Mat4* TYPE_MATRIX;
			ReferenceType* TYPE_REFERENCE;
			ScopeType* TYPE_SCOPE;
			std::string* TYPE_STRING;
			PointerType TYPE_POINTER;
			RTTIType* TYPE_RTTI;
			//Constructor
			DataType() : TYPE_UNKNOWN(NULL){}
		};

	private:

		TypeE mType; // Type of data
		bool mIsExternal; // Marks if the data is external or not (aka if Datum owns the memory or not)
		
		unsigned int mSize; //
		unsigned int mCapacity;

		DataType mData;
};


