#pragma once


#include <string>
#include <sstream>
#include <stdio.h>
#include <utility>
#include <stdlib.h>

#include "SList.h"
#include "Vector.h"
#include "HashMap.h"

#include "RTTI.h"
#include "Scope.h"

template <class AbstractProductT> class Factory
{
    // Interface (Abstract Base Class):
    public:
		typedef Factory<AbstractProductT>  SpecializedFactory;
		typedef HashMap<std::string, SpecializedFactory*, HashString> HashTable;

		//! Deconstructor
		/*!
			Virtual deconstructor
		*/
		virtual ~Factory(){};

		//! GetClassName
		/*!
			Virtual function that returns the class name that this factory creates
			\return - returns a string of the product class that this factory creates 
		*/
		virtual const std::string GetClassName(void) = 0;


	private:
		//! Create
		/*!
			Virtual function that creates the product type and returns a pointer to it
			\return - returns a pointer to the created product
		*/
        virtual AbstractProductT * Create(void) = 0;

    // Manager (Singleton):
    public:

		//! Find
		/*!
			Returns a pointer to the factory that creates the passed in product
			\param - a the product of the factory to find
			\return - returns a pointer to the created product
		*/
		static SpecializedFactory * Find(const std::string strClassName);

		//! Create
		/*!
			Returns a pointer to the factory that creates the passed in product
			\param - a the product of the factory to find
			\return - returns a pointer to the created product
		*/
		static AbstractProductT * Create(const std::string strClassName);
		
    protected:

		//! Add
		/*!
			Adds the passed in factory to the factory manager
			\param - a pointer to the factory to add
		*/
        static void Add   ( SpecializedFactory * pFactory );

		//! Remove
		/*!
			Removes the factory from the factory manager with the product that the passed in factory pointer creates
			\param - a pointer to the factory to remove
		*/
        static void Remove( SpecializedFactory * pFactory );

		//! Begin
		/*!
			Returns a hashmap iterator that points to the front of the hash map
			\return - a hashmap iterator that points to the front of the hash map
		*/
		static typename HashTable::HashMapIterator Begin() { return(sFactoryTable.Begin()); }

		//! End
		/*!
			Returns a hashmap iterator that points past the to the end of the hash map
			\return - a hashmap iterator that points past the to the end of the hash map
		*/
		static typename HashTable::HashMapIterator End(){ return(sFactoryTable.End()); }

    private:
		static HashTable sFactoryTable;
};

template<typename AbstractProductT>
Factory<AbstractProductT> * Factory<typename AbstractProductT>::Find(const std::string strClassName ) 
{
	assert( sFactoryTable.Find(strClassName) != sFactoryTable.End() );
	return(sFactoryTable.Find(strClassName)->second);
}

template<typename AbstractProductT>
AbstractProductT * Factory<typename AbstractProductT>::Create( const std::string strClassName )
{
	assert( sFactoryTable.Find(strClassName) != sFactoryTable.End() );
	return(sFactoryTable.Find(strClassName)->second->Create());
}

template<typename AbstractProductT>
void Factory<typename AbstractProductT>::Add(SpecializedFactory * pFactory ) 
{
	pair < std::string, Factory<AbstractProductT>* > newFactoryEntry( pFactory->GetClassName(), pFactory );
	sFactoryTable.Insert( newFactoryEntry );
}

template<typename AbstractProductT>
void Factory<typename AbstractProductT>::Remove(SpecializedFactory * pFactory ) 
{
	sFactoryTable.Remove(pFactory->GetClassName()); 
}

//Static Hash Map Init
template<typename AbstractProductT>
HashMap< std::string, Factory<AbstractProductT>*, HashString > Factory<AbstractProductT>::sFactoryTable(20);

//Macro
#define ConcreteFactory( AbstractProductT, ConcreteProductT )				\
	class ConcreteProductT ## Factory : Factory<AbstractProductT>			\
    {																		\
        public:																\
             ConcreteProductT ## Factory()  { Add   ( this ) ; }			\
            ~ConcreteProductT ## Factory()  { Remove( this ) ; }			\
            virtual const std::string GetClassName( void )					\
			{ return std::string(# ConcreteProductT); }						\
            virtual AbstractProductT * Create( void )						\
            {																\
                AbstractProductT * product = new ConcreteProductT();		\
                assert( product != NULL ) ;									\
                return product ;											\
            }																\
	};																		\


