#pragma once

#include <string>
#include <sstream>
#include <stdio.h>
#include <utility>

#include "RTTI.h"
#include "vec4.h"
#include "mat4.h"
#include "HashMap.h"
#include "Vector.h"
#include "Datum.h"
#include "Limits.h"

class Scope : public IClonable
{
	public:
		RTTI_DECLARATIONS(Scope, IClonable)
		
		typedef pair<std::string, Datum*> DictionaryEntry; 

		//! Constructor
		/*!
			shall create a scope of of the passed in size, defaults to 20.
			\param - inital size of the scope
		*/
		Scope(unsigned int intialSize = 20);
		
		//! Deonstructor
		/*!
			shall destroy the scope and all allocated memory
		*/
		virtual ~Scope();
		
		//! Copy Constructor
		/*!
			shall deep copy the scope
			\param - the scope to copy
		*/
		Scope(const Scope& source);
		

		//! Operator =
		/*!
			shall assign one scope to another by making a deep copy of it and all it's children.
			\param - the scope to copy from
			\return - a reference to the new scope
		*/
		Scope& operator=(const Scope& source);
		
		
		//! Operator == 
		/*!
			shall compare two scopes
			\param - the scope to compare to
			\return - a bool determining if the Scopes are equal or not
		*/
		bool operator==(const Scope& source);
		
		//! Operator != 
		/*!
			shall compare two scopes
			\param - the scope to compare to
			\return - a bool determining if the Scopes are equal or not
		*/
		bool operator!=(const Scope& source);
		
		//! Operator[] 
		/*!
			shall return a reference to the data witht the passed name, returns null if not found
			\param - the name of the datum inside this scope to return
			\return - a bool determining if the Scopes are equal or not
		*/
		Datum& operator[](const string index);
		
		//! Operator[] 
		/*!
			shall return a reference to the data at the passed index, asserts if invalid index
			\param - the index of the datum inside this scope to return
			\return - a reference to the found datum
		*/
		Datum& operator[](const unsigned int index) const;

		//! Size 
		/*!
			shall return the size of the scope
			\return - a unsigned int that is the number if table entries in the scope
		*/
		const unsigned int Size() const;
		
		//! Find
		/*!
			shall return a reference to the datum with the given name at this scope level, returns null if not found
			\param - the name of the datum to search for
			\return - a reference to the found datum
		*/
		Datum* Find(const std::string name);

		//! Search
		/*!
			shall return a reference to the datum with the given name in this scope or any of it's ancestors, returns null if not found
			\param - the name of the datum to search for
			\return - a reference to the found datum
		*/
		Datum* Search(const std::string name, Scope** scopePointer = NULL);


		//! Append
		/*!
			shall create a new datum inside the scope with the given name, or returns the datum with the name already if it exists.
			\param - the name of the datum to create/search for
			\return - the datum with the passed name, either found ot created
		*/
		Datum& Append(const std::string name);
		
		//! AppendScope
		/*!
			shall return the scope at the passed index inside the datum with the passed name, if no such datum exists it will
			create a new datum and a new scope.  if the index is equal to the size of the datum it create a new scope inside the datum.
			If it is past that it will assert. If the datum is not of type scope it will assert.  Also if the proposed child will create
			a circular chain the program will assert.

			\param - the name of the datum to create/search for
			\param - the index into the datum of the scope to return
			\return - the found or created scope
		*/
		Scope& AppendScope(const std::string name, const unsigned int index = 0);
		
		//! Adopt
		/*!
			shall adopt the passed scope child inserting it into a new datum with the passed name.  
			The new datum will be inserted inside the scope's table at the passed index.  Will assert if the passed name
			already exists.
			\param - the child to adopt.
			\param - the name of the datum to create.
			\param - the index into scope's table to insert the new datum at
		*/
		void Adopt(Scope* childToAdopt, std::string nameOfChild, unsigned int indexToInsertAt);

		//! GetParent
		/*!
			shall get the parent of this scope
			\return - returns a pointer to the parent of this scope
		*/
		Scope* GetParent() const;

		//! CopyValues
		/*!
			shall copy the values from one scope into another.  Does not clear the existing values in the scope, but will
			override them if the override flag is passed in as true.  Will append copies of the source scope's datums onto 
			this scope.
			\param - the scope to copy
			\param - a bool determining if datum's with the same name should be overriden.
		*/
		void CopyValues(const Scope& source, bool overrideConflicts = false);


		//! Clear
		/*!
			shall clear the scope and delete all allocated memory.
		*/
		void Clear();


		//! ReverseFind
		/*!
		*/
		bool ReverseFind(const Scope* scopeReference, std::string& stringOutputReference);


		//! GetMemberName
		/*!
			shall get the name the entry in the scope at that index
			\param - the index of the datum in this scope table whos name to return
			\return - the corresponding name
		*/
		std::string GetMemberName(const unsigned int index) const;

		//! GetScopeIndex
		/*!
			shall get the index of the given scope if it exists in this scope, if it doesn't exist then it will return max int.
			\param - a reference to the child scope who's index to search for 
			\return - the index of the passed scope
		*/
		const unsigned int GetScopeIndex(Scope* scopeReference);
		
		//! GetEntryIndex
		/*!
			shall get the index of the given datum if it exists in this scope, if it doesn't exist then it will return max int.
			\param - a string of the datum to loom for
			\return - the index of the found datum
		*/
		const unsigned int GetEntryIndex(std::string name);
		
		//! GetEntry
		/*!
			shall get the datum if it exists in this scope of the passed in index, if it doesn't exist then it will return null.
			\param - the index of the datum to return
			\return - a pointer to the datum at that index
		*/
		Datum * GetEntry(unsigned int index);


		//! RemoveEntry
		/*!
			shall remove the entry associated with the passed in name
			\param - a string of the entry to remove
		*/
		void RemoveEntry(std::string name);


		//! Clone
		/*! Clones a Scope table using a virtual constructor.
			\param strCloneName - name parent give sto the cloned Scope
			\param index - index within the values array associated with the given name to assign to clone
		*/
		/* virtual */ 
		//Datum::IClonable * Scope::Clone( const char * strCloneName , unsigned int index )
		virtual IClonable * Clone();


		//! GetOwnName
		/*! Gets this scopes own name
			\return - return's this scope's name in it's parent's entry table
		*/
		std::string GetOwnName();

		//! RemoveScopeEntry
		/*!
			shall remove the corresponding entry from the the datum with the passed name at the corresponding index.
			Will kill the scope at that index if the datum is of type scope and if the kill flag is true.
			\param - the name of the datum
			\param - the index into the datum
			\param - a bool determining if the scope should be destroyed or not
		*/
		void RemoveScopeEntry(std::string name, const unsigned int index, bool killScope = false);

	protected:
		//! BecomeOrphan
		/*!
			shall become a orphan, will destory this scope's connect to it's parent and it's parent's connection to it.
		*/
		void BecomeOrphan();

	private:
		Scope* mParent;

		Vector<DictionaryEntry*> mTable;
		HashMap<std::string, Datum*, HashString> mData;
		
};
