#pragma once

#include <iostream>
#include <assert.h>

#include "SList.h"

/*!
	1. If you choose to use a fixed size stack you then need to pass the capacity of the stack.

	2. The class does not include an iterator because a stack by definition is fifo (first in first out).  
		It wouldn't be a stack if it had an iterator.

	3.  I choose return nothing because you already have the top method.  This also eliminates the issue that occurs
		if the user of the stack attempts to pop when nothing exists.

	4.  The stl is an container adapter that allows you to set what the internal storage of elements is handled.  This allows
		to choose a specific internal container for preformance reasons.   

	5.  I choose to use SList because a singlely linked list is pretty much a stack.  Reusing the SList and creating a wrapper
		around it tests my SList more and recreating code that I already have is just a waste of time.
*/

template<typename TemplatedClass>
class Stack
{

public:
	Stack();
	~Stack();
	Stack(const Stack& source);
	Stack<TemplatedClass> &operator=(const Stack& source);
	void Push(const TemplatedClass & newData);
	void Pop();
	TemplatedClass & Top();
	const TemplatedClass & Top() const;
	unsigned int Size() const;
	bool IsEmpty() const;
	void Clear();


private:
	SList<typename TemplatedClass> myStack;
};

//! Constructor
/*!
	shall initialize the list to be empty.
*/
template<typename TemplatedClass>
Stack<typename TemplatedClass>::Stack()
{
}

//! Deconstructor
/*!
	shall deconstruct the stack.
*/
template<typename TemplatedClass>
Stack<typename TemplatedClass>::~Stack()
{
	myStack.Clear();
}

//! Copy Constructor
/*!
	A copy constructor
*/
template<typename TemplatedClass>
Stack<typename TemplatedClass>::Stack(const Stack& source)
{
	operator=(source);
}

//! Operator=
/*!
	shall deep-copy the list given as argument into this list.
	\param - the vector to copy from
*/
template<typename TemplatedClass>
Stack<typename TemplatedClass>& Stack<typename TemplatedClass>::operator=(const Stack& source)
{
	Clear(); // We need a empty list to start copying
	myStack = source.myStack;
	return(*this);
}

//! Push
/*!
	shall push the given item to the front of the stack.
	\param - the data to add to the vector
*/
template<typename TemplatedClass>
void Stack<typename TemplatedClass>::Push(const TemplatedClass & newData)
{
	myStack.PushFront(newData);
}

//! Stack
/*!
	shall pop the off the top of the stack
*/
template<typename TemplatedClass>
void Stack<typename TemplatedClass>::Pop()
{
	if(myStack.Size() > 0){
		myStack.PopFront();
	}
}

//! Top
/*!
	shall get the data on the top of the stack
	\return - the data on the top of the stack as a reference
*/
template<typename TemplatedClass>
TemplatedClass & Stack<typename TemplatedClass>::Top()
{
	return(myStack.Front());
}

//! Top
/*!
	shall get the data on the top of the stack
	\return - the data on the top of the stack as a const reference
*/
template<typename TemplatedClass>
const TemplatedClass & Stack<typename TemplatedClass>::Top() const
{
	return(myStack.Front());
}

//! Size
/*!
	shall get size of the stack
	\return - the size of the stack
*/
template<typename TemplatedClass>
unsigned int Stack<typename TemplatedClass>::Size() const
{
	return(myStack.Size());
}

//! IsEmpty
/*!
	shall return a bool if the stack is empty or not
	\return - a bool indicating if the stack is empty or not
*/
template<typename TemplatedClass>
bool Stack<typename TemplatedClass>::IsEmpty() const
{
	if(myStack.Size() > 0)
		return(false);
	else
		return(true);

}

//! Clear
/*!
	shall clear the stack
*/
template<typename TemplatedClass>
void Stack<typename TemplatedClass>::Clear()
{
	myStack.Clear();
}