#pragma once

#include <iostream>
#include <assert.h>
#include <math.h>

#define MAX(a, b) (((a) > (b)) ? (a) : (b))

template<typename TemplatedClass>
class Vector
{
	public:
		//! Iterator Class
		class VectorIterator
		{
			friend class Vector;
			public:
				~VectorIterator();
				VectorIterator(const VectorIterator & source);
				VectorIterator &operator=(const VectorIterator & source);
				bool operator==(const VectorIterator & otherIterator) const;
				bool operator!=(const VectorIterator & otherIterator) const;
				VectorIterator operator++(int);
				VectorIterator & operator++();
				VectorIterator operator--(int);
				VectorIterator & operator--();
				TemplatedClass & operator*();
			private:
				VectorIterator();
				unsigned int location;
				Vector* parent;
			
		};		

		Vector();
		~Vector();
		Vector(const Vector& source);
		Vector<TemplatedClass> &operator=(const Vector& source);
		const TemplatedClass& operator[](const unsigned int i) const;
		TemplatedClass& operator[](const unsigned int i);
		void PushBack(const TemplatedClass & newData);
		void PopBack();
		bool Empty() const;
		const TemplatedClass & Front() const;
		TemplatedClass & Front();
		const TemplatedClass & Back() const;
		TemplatedClass & Back();
		const unsigned int Size() const;
		VectorIterator Begin(); 
		VectorIterator End(); 
		TemplatedClass& At(const unsigned int returnedLocation);
		void Reserve(const unsigned int newCapacity);
		VectorIterator Find(const TemplatedClass desiredItem);
		void Clear();
		const unsigned int Capacity() const;
		void Resize(const unsigned int newCapacity);
		void InsertBefore(const TemplatedClass & newData, const unsigned int index);
		void Remove(const unsigned int index);
	private:
		TemplatedClass* vectorData;
		unsigned int size; // Size of the list
		unsigned int capacity;
};



//**********************************************
//Vector Iterator
//**********************************************

//! Constructor
/*!
	A constructor that creates a new iterator.
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::VectorIterator::VectorIterator()
{
}

//! Deconstructor
/*!
	A deconstructor that assures there are no memory leaks.  
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::VectorIterator::~VectorIterator()
{
}

//! Copy Constructor
/*!
	A copy constructor
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::VectorIterator::VectorIterator(const VectorIterator& source)
{
	operator=(source);
}

//! = operator overload
/*!
	Overloads the = operator
	\param - the iterator to set to
	\return - reference to the source iterator
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator & Vector<typename TemplatedClass>::VectorIterator::operator=(const VectorIterator& source)
{
	parent = source.parent;
	location = source.location;
	return(*this);
}


//! == operator overload
/*!
	Overloads the == operator
	\param - the iterator to compare to
	\return - a bool stating if the iterators are equal or not
*/
template<typename TemplatedClass>
bool Vector<typename TemplatedClass>::VectorIterator::operator==(const VectorIterator & otherIterator) const
{
	if(parent == otherIterator.parent){
		if(location == otherIterator.location)
			return(true);
	}
	return(false);
}

//! != operator overload
/*!
	Overloads the == operator
	\param - the iterator to compare to
	\return - a bool stating if the iterators are equal or not
*/
template<typename TemplatedClass>
bool Vector<typename TemplatedClass>::VectorIterator::operator!=(const VectorIterator & otherIterator) const
{
	return(!operator==(otherIterator));
}

//! Postfix operator
/*!
	Overloads the ++ postfix incrementaor operator
	\param - a dummy  variable
	\return - a iterator reference that is pointing at the current spot then increases the iterator
	\asserts that the iterator is not trying to go past end
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator Vector<typename TemplatedClass>::VectorIterator::operator++(int)
{
	assert(location < (parent->Size()));
	VectorIterator tempCopy = (*this);
	location++;
	return(tempCopy);
}

//! Prefix operator
/*!
	Overloads the ++ prefix incrementaor operator
	\return - a iterator reference that is pointing at the next spot
	\asserts that the iterator is not trying to go past end
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator & Vector<typename TemplatedClass>::VectorIterator::operator++()
{
	assert(location < (parent->Size()));
	location++;
	return(*this);
}

//! Postfix operator
/*!
	Overloads the -- postfix incrementaor operator
	\param - a dummy  variable
	\return - a iterator reference that is pointing at the current spot then decreases the iterator
	\asserts - that the iterator is not trying to go past the front element
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator Vector<typename TemplatedClass>::VectorIterator::operator--(int)
{
	assert(location > 0);
	VectorIterator tempCopy = (*this);

	// Call prefix
	operator--();

	return(tempCopy);
}

//! Prefix operator
/*!
	Overloads the -- prefix incrementaor operator
	\returns - a iterator reference that is pointing at the eariler spot
	\asserts - that the iterator is not trying to go past the front element
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator & Vector<typename TemplatedClass>::VectorIterator::operator--()
{
	assert(location > 0);
	location--;
	return(*this);
}

//! Dereference operator (operator*)
/*!
	Overloads the -- prefix incrementaor operator
	\return - a reference to the data
*/
template<typename TemplatedClass>
TemplatedClass & Vector<typename TemplatedClass>::VectorIterator::operator*()
{
	assert((location >= 0) && (location < parent->Size()));
	return(parent->vectorData[location]);
}

//**********************************************
//Vector Class
//**********************************************

//! Constructor
/*!
	shall initialize the list to be empty.
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::Vector() : size(0), capacity(0), vectorData(NULL)
{
}

//! Copy Constructor
/*!
	shall deep-copy the list given as argument into this list.
	\param - the vector to copy from
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::Vector(const Vector& source)  : size(0), capacity(0), vectorData(NULL)
{
	size = 0;
	capacity = 0;
	operator=(source);
}

//! Operator=
/*!
	shall deep-copy the list given as argument into this list.
	\param - the vector to copy from
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>& Vector<typename TemplatedClass>::operator=(const Vector& source)
{
	Clear(); // We need a empty list to start copying

	for(unsigned int i = 0; i < source.Size(); i++){
		PushBack(source.vectorData[i]);
	}
	return(*this);
}

//! Deconstructor
/*!
	shall initialize the list to be empty.
*/
template<typename TemplatedClass>
Vector<typename TemplatedClass>::~Vector()
{
	Clear();
	if(capacity > 0){
		if(vectorData){
			//delete vectorData;
			vectorData = NULL;
		}
	}
	capacity = 0;
}


//! [] operator overload
/*!
	Overloads the [] operator
	\param - the location of the data to return
	\return - a reference to the data at that location
*/
template<typename TemplatedClass>
const TemplatedClass& Vector<typename TemplatedClass>::operator[](const unsigned int i) const
{
	assert(i < capacity);
	return(vectorData[i]);
}

//! [] operator overload
/*!
	Overloads the [] operator
	\param - the location of the data to return
	\return - a reference to the data at that location
*/
template<typename TemplatedClass>
TemplatedClass& Vector<typename TemplatedClass>::operator[](const unsigned int i)
{
	assert(i < size);
	return(vectorData[i]);
}

//! PushBack
/*!
	shall append given item after the end of the list.  This must increase array capacity if necessary.
	\param - the data to add to the vector
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::PushBack(const TemplatedClass & newData)
{
	size++; // Increase the size
	if(size > capacity){ // Do a check to see if resize is needed
		if(capacity == 0)
			Reserve(1);
		else
			Reserve((MAX(1,(capacity / 2))) + capacity);
	}
	//vectorData[size - 1] = newData;
	new (&vectorData[size - 1]) TemplatedClass(newData);
}

//! PopBack
/*!
	shall remove the last item in the container, but should not reduce the capacity of the container.  This must destruct the item deleted. 
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::PopBack()
{
	if(size > 0){
		vectorData[size - 1].~TemplatedClass();
		size--;
	}
}

//! Empty
/*!
	shall return a bool indicating whether the list contains any items.
	\return - a bool that determines if the list is empty or not
*/
template<typename TemplatedClass>
bool Vector<typename TemplatedClass>::Empty() const
{
	if(size > 0)
		return(false);
	return(true);
}

//! Front
/*!
	shall return a const copy of first item in the container.
	\return - a const copy of the data at the front of the list
*/
template<typename TemplatedClass>
const TemplatedClass & Vector<typename TemplatedClass>::Front() const
{
	assert(size > 0);
	return(vectorData[0]);
}

//! Front
/*!
	shall return a copy of first item in the container.
	\return - a copy of the data at the front of the list
*/
template<typename TemplatedClass>
TemplatedClass & Vector<typename TemplatedClass>::Front()
{
	assert(size > 0);
	return(vectorData[0]);
}

//! Back
/*!
	shall return a const copy of last item in the container.
	\return - a const copy the data at the end of the list
*/
template<typename TemplatedClass>
const TemplatedClass & Vector<typename TemplatedClass>::Back() const
{
	assert(size > 0);
	return(vectorData[size - 1]);
}

//! Back
/*!
	shall return a copy of last item in the container.
	\return - a copy of the data at the end of the list
*/
template<typename TemplatedClass>
TemplatedClass & Vector<typename TemplatedClass>::Back()
{
	assert(size > 0);
	return(vectorData[size - 1]);
}

//! Size
/*!
	shall return the number of items in the container as an unsigned int.  Hint: Either End or Size can be efficient, and End is probably called more, so what does that imply about how to represent your container?
	\return - an unsigned int that is the size of the list
*/
template<typename TemplatedClass>
const unsigned int Vector<typename TemplatedClass>::Size() const
{
	return(size);
}

//! Begin
/*!
	shall return an Iterator (see below) pointing to the first element in the container.
	\return - a iterator that points at the front of the vector
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator Vector<TemplatedClass>::Begin()
{
	assert(size > 0);
	VectorIterator newIterator;
	newIterator.location = 0;
	newIterator.parent = (this);
	return(newIterator);
}

//! End
/*!
	shall return an Iterator (see below) pointing to the last element in the container.
	\return - a iterator that points past the end of the list
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator Vector<TemplatedClass>::End()
{
	assert(size > 0);
	VectorIterator newIterator;
	newIterator.location = size;
	newIterator.parent = (this);
	return(newIterator);
}


//! At
/*!
	shall take an integer index and return a reference to the item at the given index.
	\param - the location of the data to return
	\return - data at the desired location
*/
template<typename TemplatedClass>
TemplatedClass& Vector<typename TemplatedClass>::At(const unsigned int returnedLocation)
{
	assert((returnedLocation >= 0) && (returnedLocation < size));
	return((*this)[returnedLocation]);
}

//! Reserve
/*!
	shall take an integer indicating the minimum capacity to reserve for the array, i.e. the number of elements allocated for the array  whether or not they are occupied.  Reserve may not leak memory.  If memory was previously allocated, Reserve shall preserve the contents of the former array.  Reserve may not delete items in the container.  When allocating the array, it is preferable to use C-style malloc (i.e. not new), regardless of the container type.
	\param - the amount of data to attempt to reserve
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::Reserve(const unsigned int newCapacity)
{
	if(newCapacity > capacity) {
		unsigned int oldCapacity = capacity;
		capacity = newCapacity;
		if(oldCapacity == 0){ // Special Case
			TemplatedClass* newVectorData; // Create a pointer to the new allocated data
			
			newVectorData = new TemplatedClass[capacity]();
			//newVectorData = (TemplatedClass*)malloc(sizeof(TemplatedClass) * capacity); // Malloc the new space
			
			
			vectorData = newVectorData; // point the Vector's vectorData to the newVectorData
		}
		else{
			TemplatedClass* newVectorData; // Create a pointer to the new allocated data
			
			//newVectorData = (TemplatedClass*)malloc(sizeof(TemplatedClass) * capacity); // Malloc the new space
			
			newVectorData = new TemplatedClass[capacity]();

			memcpy( (newVectorData), (vectorData), oldCapacity * sizeof(TemplatedClass) ); // need to multiple the oldCapacity by 4 to get the correct number of bytes, 4 bytes per address )
			
			//delete vectorData; // Remove the old vector data
			delete[] vectorData; // Remove the old vector data
			vectorData = newVectorData; // point the Vector's vectorData to the newVectorData
		}
	}
}

//! Find
/*!
	Finds the desired item
	\param - the data to searh for
	\return - at vector pointing at desired data
*/
template<typename TemplatedClass>
typename Vector<TemplatedClass>::VectorIterator Vector<typename TemplatedClass>::Find(const TemplatedClass desiredItem)
{
	for(VectorIterator myIterator = Begin(); myIterator != End(); myIterator++){
		if((*myIterator) == desiredItem){
			return(myIterator);
		}
	}
	return(End());
}

//! Clear
/*!
	shall empty the container.  This must destruct all items deleted.
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::Clear()
{
	if(capacity > 0){
		for(unsigned int i = 0; i < capacity; i++){
			vectorData[i].~TemplatedClass();
			//delete vectorData[i];
		}
	}
	//delete[] vectorData;
	size = 0;
}

//! Capacity
/*!
	shall return the current capacity of the vector
	\return - an unsigned int that is the capacity
*/
template<typename TemplatedClass>
const unsigned int Vector<typename TemplatedClass>::Capacity() const
{
	return(capacity);
}

//! Resize
/*!
	shall resize the vector, grow or shrink
	\param - an unsigned int that is new capacity
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::Resize(const unsigned int newCapacity)
{
	unsigned int copyCapacity = capacity;
	if(newCapacity < capacity)
		copyCapacity = newCapacity;

	TemplatedClass* oldVectorData = vectorData;
	TemplatedClass* newVectorData;
	
	newVectorData = new TemplatedClass[capacity]();
	memcpy( newVectorData , oldVectorData, copyCapacity * sizeof(TemplatedClass) ); // Copy the data

	for(unsigned int i = copyCapacity ; i < capacity; i++){
		oldVectorData[i] = 0;
	}
	delete[] oldVectorData;
	
	//Set the vectorData to the new array
	vectorData = newVectorData;

	if(size > newCapacity)
		size = newCapacity;
	capacity = newCapacity;

	oldVectorData = 0;
	newVectorData = 0;
	if(newCapacity == 0)
		vectorData = 0;
}

//! Insert
/*!
	shall insert the passed data at the corresponding location in the vector, moving everything back
	\param - the data to insert
	\param - an unsigned int that is location to insert at
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::InsertBefore(const TemplatedClass & newData, const unsigned int index)
{
	if(index > capacity)
		Reserve(index);

	//unsigned int start = (size - 1);
	if(size == 0){
		PushBack(newData);
	}
	else{
		for(unsigned int i = (size - 1); i >= index; i--){
			vectorData[i] = vectorData[i - 1];
		}
		vectorData[index] = newData;
		size++;
	}
}

//! Remove
/*!
	shall remove the element at the specified index
	\param - an unsigned int that is location to remove at
*/
template<typename TemplatedClass>
void Vector<typename TemplatedClass>::Remove(const unsigned int index)
{
	if(index < size){
		for(unsigned int i = index; i < size; i++){
			vectorData[i] = vectorData[i + 1];
		}
	}
	PopBack();
}
