/*! \file mat4.h

    \brief 4x4 matrix class

    \author Copyright 2005-2006 UCF/FIEA/MJG; All rights reserved.
*/
#ifndef MAT4_H
#define MAT4_H

#include "util.h"
#include "macros.h"
#include "float.h"
#include "vec4.h"
#include "vec3.h"

// Macros --------------------------------------------------------------

//! \brief Cast an array of floats to a Mat4
#define M4(m4) (*(Mat4*)m4)

// Types --------------------------------------------------------------

/*! \brief Utility class for 4x4 matrices

    \todo   Rewrite this class so that the matrix is composed of canonical vectors,
            such that matrix-vector transformation has the form
                v' = M_x * v_x, M_y * v_y, M_z * v_z, M_w * v_w
            where * are scalar products.
            This rewrite would be tantamount to representing M as4 4-vectors,
            where M_x == "right", M_y = "forward", M_z = "up", M_z = "position".
            This rewrite would allow several methods of this class to be written
            in terms of Vec4 operations, which would mean that any use of
            vectorized hardware would also accelerate this class.

            This would also mean that the accessors for Forward, Right, Up and Position
            could return by reference instead of by value, which could
            provide additional computational efficiency.

*/
struct Mat4
{
    public:
        inline Mat4() {}
        inline Mat4(float m0, float m1, float m2, float m3,
                    float m4, float m5, float m6, float m7,
                    float m8, float m9, float ma, float mb,
                    float mc, float md, float me, float mf)
        {
            m[0][0] = m0; m[0][1] = m1; m[0][2] = m2; m[0][3] = m3;
            m[1][0] = m4; m[1][1] = m5; m[1][2] = m6; m[1][3] = m7;
            m[2][0] = m8; m[2][1] = m9; m[2][2] = ma; m[2][3] = mb;
            m[3][0] = mc; m[3][1] = md; m[3][2] = me; m[3][3] = mf;
        }
        // assignment operators
        inline Mat4 & operator += (const Mat4 & rhs)    { (*this) = (*this) + rhs; return * this; }
        inline Mat4 & operator -= (const Mat4 & rhs)    { (*this) = (*this) + (-1.0f * rhs); return * this; }
        inline Mat4 & operator *= (const float& f)    { (*this) = (*this) * f; return * this; }
        //inline Mat4 & operator /= (float f)           { (*this) = (*this) / rhs; return * this; }

        // unary operators
        inline Mat4 operator + () const { return * this; }
        inline Mat4 operator - () const { return -1.0f * (*this); }
        //Inverse Matrix
        inline Mat4 operator ~ () const 
        { 
            Mat4 inverse;
            float det = this->Determinant();
            ASSERT (det != 0);

            return Mat4(
            (m[1][1]*m[2][2]*m[3][3] + m[1][2]*m[2][3]*m[3][1] + m[1][3]*m[2][1]*m[3][2] - m[1][1]*m[2][3]*m[3][2] - m[1][2]*m[2][1]*m[3][3] - m[1][3]*m[2][2]*m[3][1]) / det,
            (m[0][1]*m[2][3]*m[3][2] + m[0][2]*m[2][1]*m[3][3] + m[0][3]*m[2][2]*m[3][1] - m[0][1]*m[2][2]*m[3][3] - m[0][2]*m[2][3]*m[3][1] - m[0][3]*m[2][1]*m[3][2]) / det,
            (m[0][1]*m[1][2]*m[3][3] + m[0][2]*m[1][3]*m[3][1] + m[0][3]*m[1][3]*m[3][2] - m[0][1]*m[1][3]*m[3][2] - m[0][2]*m[1][1]*m[3][3] - m[0][3]*m[1][2]*m[3][1]) / det,
            (m[0][1]*m[1][3]*m[2][2] + m[0][2]*m[1][1]*m[2][3] + m[0][3]*m[1][2]*m[2][1] - m[0][1]*m[1][2]*m[2][3] - m[0][2]*m[1][3]*m[2][1] - m[0][3]*m[1][1]*m[2][2]) / det,
            (m[1][0]*m[2][3]*m[3][2] + m[1][2]*m[2][0]*m[3][3] + m[1][3]*m[2][2]*m[3][0] - m[1][0]*m[2][2]*m[3][3] - m[1][2]*m[2][3]*m[3][0] - m[1][3]*m[2][0]*m[3][2]) / det,
            (m[0][0]*m[2][2]*m[3][3] + m[0][2]*m[2][3]*m[3][0] + m[0][3]*m[2][0]*m[3][2] - m[0][0]*m[2][3]*m[3][2] - m[0][2]*m[2][0]*m[3][3] - m[0][3]*m[2][2]*m[3][0]) / det,
            (m[0][0]*m[1][3]*m[3][2] + m[0][2]*m[1][0]*m[3][3] + m[0][3]*m[1][2]*m[3][0] - m[0][0]*m[1][2]*m[3][3] - m[0][2]*m[1][3]*m[3][0] - m[0][3]*m[1][0]*m[3][2]) / det,
            (m[0][0]*m[1][2]*m[2][3] + m[0][2]*m[1][3]*m[2][0] + m[0][3]*m[1][0]*m[2][2] - m[0][0]*m[1][3]*m[2][2] - m[0][2]*m[1][0]*m[2][3] - m[0][3]*m[1][2]*m[2][0]) / det,
            (m[1][0]*m[2][1]*m[3][3] + m[1][1]*m[2][3]*m[3][0] + m[1][3]*m[2][0]*m[3][1] - m[1][0]*m[2][3]*m[3][1] - m[1][1]*m[2][0]*m[3][3] - m[1][3]*m[2][1]*m[3][0]) / det,
            (m[0][0]*m[2][3]*m[3][1] + m[0][1]*m[2][0]*m[3][3] + m[0][3]*m[2][1]*m[3][0] - m[0][0]*m[2][1]*m[3][3] - m[0][1]*m[2][3]*m[3][0] - m[0][3]*m[2][0]*m[3][1]) / det,
            (m[0][0]*m[1][1]*m[3][3] + m[0][1]*m[1][3]*m[3][0] + m[0][3]*m[1][0]*m[3][1] - m[0][0]*m[1][3]*m[3][1] - m[0][1]*m[1][0]*m[3][3] - m[0][3]*m[1][1]*m[3][0]) / det,
            (m[0][0]*m[1][3]*m[2][1] + m[0][1]*m[1][0]*m[2][3] + m[0][3]*m[1][1]*m[2][0] - m[0][0]*m[1][1]*m[2][3] - m[0][1]*m[1][3]*m[2][0] - m[0][3]*m[1][0]*m[2][1]) / det,
            (m[1][0]*m[2][2]*m[3][1] + m[1][1]*m[2][0]*m[3][2] + m[1][2]*m[2][1]*m[3][0] - m[1][0]*m[2][1]*m[3][2] - m[1][1]*m[2][2]*m[3][0] - m[1][2]*m[2][0]*m[3][1]) / det,
            (m[0][0]*m[2][1]*m[3][2] + m[0][1]*m[2][2]*m[3][0] + m[0][2]*m[2][1]*m[3][1] - m[0][0]*m[2][2]*m[3][1] - m[0][1]*m[2][0]*m[3][2] - m[0][2]*m[2][1]*m[3][0]) / det,
            (m[0][0]*m[1][2]*m[3][1] + m[0][1]*m[1][0]*m[3][2] + m[0][2]*m[1][1]*m[3][0] - m[0][0]*m[1][1]*m[3][2] - m[0][1]*m[1][2]*m[3][0] - m[0][2]*m[1][0]*m[3][1]) / det,
            (m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] + m[0][2]*m[1][0]*m[2][1] - m[0][0]*m[1][2]*m[2][1] - m[0][1]*m[1][0]*m[2][2] - m[0][2]*m[1][1]*m[2][0]) / det);
        } 

        inline float Determinant () const
        {
            return (m[0][3] * m[1][2] * m[2][1] * m[3][0] - m[0][2] * m[1][3] * m[2][1] * m[3][0] - m[0][3] * m[1][1] * m[2][2] * m[3][0] + m[0][1] * m[1][3] * m[2][2] * m[3][0] +
                    m[0][2] * m[1][1] * m[2][3] * m[3][0] - m[0][1] * m[1][2] * m[2][3] * m[3][0] - m[0][3] * m[1][2] * m[2][0] * m[3][1] + m[0][2] * m[1][3] * m[2][0] * m[3][1] +
                    m[0][3] * m[1][0] * m[2][2] * m[3][1] - m[0][0] * m[1][3] * m[2][2] * m[3][1] - m[0][2] * m[1][0] * m[2][3] * m[3][1] + m[0][0] * m[1][2] * m[2][3] * m[3][1] +
                    m[0][3] * m[1][1] * m[2][0] * m[3][2] - m[0][1] * m[1][3] * m[2][0] * m[3][2] - m[0][3] * m[1][0] * m[2][1] * m[3][2] + m[0][0] * m[1][3] * m[2][1] * m[3][2] +
                    m[0][1] * m[1][0] * m[2][3] * m[3][2] - m[0][0] * m[1][1] * m[2][3] * m[3][2] - m[0][2] * m[1][1] * m[2][0] * m[3][3] + m[0][1] * m[1][2] * m[2][0] * m[3][3] +
                    m[0][2] * m[1][0] * m[2][1] * m[3][3] - m[0][0] * m[1][2] * m[2][1] * m[3][3] - m[0][1] * m[1][0] * m[2][2] * m[3][3] + m[0][0] * m[1][1] * m[2][2] * m[3][3]);
        }

        // binary operators
        inline Mat4 operator + (const Mat4& rhs) const
        {
            return Mat4(m[0][0] + rhs.m[0][0], m[0][1] + rhs.m[0][1], m[0][2] + rhs.m[0][2], m[0][3] + rhs.m[0][3],
                        m[1][0] + rhs.m[1][0], m[1][1] + rhs.m[1][1], m[1][2] + rhs.m[1][2], m[1][3] + rhs.m[1][3],
                        m[2][0] + rhs.m[2][0], m[2][1] + rhs.m[2][1], m[2][2] + rhs.m[2][2], m[2][3] + rhs.m[2][3],
                        m[3][0] + rhs.m[3][0], m[3][1] + rhs.m[3][1], m[3][2] + rhs.m[3][2], m[3][3] + rhs.m[3][3]);
        }


        inline Mat4 operator * (const float& f) const
        {
            return Mat4(m[0][0] * f, m[0][1] * f, m[0][2] * f, m[0][3] * f,
                        m[1][0] * f, m[1][1] * f, m[1][2] * f, m[1][3] * f,
                        m[2][0] * f, m[2][1] * f, m[2][2] * f, m[2][3] * f,
                        m[3][0] * f, m[3][1] * f, m[3][2] * f, m[3][3] * f);
        }


        // inline Mat4 operator / (float f) const                   { return f * (~(*this)); }

        inline Mat4 operator - (const Mat4& rhs) const              { return (*this) + (-1.0f * rhs); }

        inline Mat4 operator / (const float& f) const             { return (*this) * (1.0f / f); }

        friend Mat4 operator * (const float& f, const Mat4& m4)   { return m4 * f; }


        /*! \brief Transform a vector by the transpose of this matrix
            \param v4 - Vector to transform
        */
        inline Vec4 operator* (const Vec4& v4) const
        {
            return Vec4(v4.x * m[0][0] + v4.y * m[0][1] + v4.z * m[0][2] + v4.w * m[0][3],
                        v4.x * m[1][0] + v4.y * m[1][1] + v4.z * m[1][2] + v4.w * m[1][3],
                        v4.x * m[2][0] + v4.y * m[2][1] + v4.z * m[2][2] + v4.w * m[2][3],
                        v4.x * m[3][0] + v4.y * m[3][1] + v4.z * m[3][2] + v4.w * m[3][3]);
        }

        /*! \brief Transform a vector by a matrix
            \param v4 - Vector to transform
            \param m4 - Transformation matrix
        */
        inline static friend Vec4 operator* (const Vec4& v4, const Mat4& m4)
        {
            return Vec4(v4.x * m4.m[0][0] + v4.y * m4.m[1][0] + v4.z * m4.m[2][0] + v4.w * m4.m[3][0],
                        v4.x * m4.m[0][1] + v4.y * m4.m[1][1] + v4.z * m4.m[2][1] + v4.w * m4.m[3][1],
                        v4.x * m4.m[0][2] + v4.y * m4.m[1][2] + v4.z * m4.m[2][2] + v4.w * m4.m[3][2],
                        v4.x * m4.m[0][3] + v4.y * m4.m[1][3] + v4.z * m4.m[2][3] + v4.w * m4.m[3][3]);
        }


        /*! \brief Transform a vector by the upper 3x3 (rotation and scale part) of a transformation matrix.
            \todo Rename to TransformNormalByTranspose
        */
        inline static Vec4 TransformNormal(const Vec4& v4, const Mat4& m4)
        {
            return Vec4(v4.x * m4.m[0][0] + v4.y * m4.m[0][1] + v4.z * m4.m[0][2],
                        v4.x * m4.m[1][0] + v4.y * m4.m[1][1] + v4.z * m4.m[1][2],
                        v4.x * m4.m[2][0] + v4.y * m4.m[2][1] + v4.z * m4.m[2][2],
                        0.0f);
        }


        /*! \brief Transform a vector by the transpose of the upper 3x3 (rotation and scale part) of a transformation matrix.
            \todo Rename to TransformNormal
        */
        inline static Vec4 TransformNormalByTranspose(const Vec4& v4, const Mat4& m4)
        {
            return Vec4(v4.x * m4.m[0][0] + v4.y * m4.m[1][0] + v4.z * m4.m[2][0],
                        v4.x * m4.m[0][1] + v4.y * m4.m[1][1] + v4.z * m4.m[2][1],
                        v4.x * m4.m[0][2] + v4.y * m4.m[1][2] + v4.z * m4.m[2][2],
                        0.0f);
        }


        /*! \brief Transform a vector by the transpose of the upper 3x3 (rotation and scale part) of a transformation matrix.
            \todo Rename to TransformNormal
        */
        inline static Vec3 TransformNormalByTranspose(const Vec3& v3, const Mat4& m4)
        {
            return Vec3(v3.x * m4.m[0][0] + v3.y * m4.m[1][0] + v3.z * m4.m[2][0],
                        v3.x * m4.m[0][1] + v3.y * m4.m[1][1] + v3.z * m4.m[2][1],
                        v3.x * m4.m[0][2] + v3.y * m4.m[1][2] + v3.z * m4.m[2][2]);
        }


        /*! \brief Transform a vector by the upper 3x3 (rotation and scale part) of a transformation matrix.
            \todo Rename to TransformNormalByTranspose
        */
        inline static Vec3 TransformNormal(const Vec3& v3, const Mat4& m4)
        {
            return Vec3(v3.x * m4.m[0][0] + v3.y * m4.m[0][1] + v3.z * m4.m[0][2],
                        v3.x * m4.m[1][0] + v3.y * m4.m[1][1] + v3.z * m4.m[1][2],
                        v3.x * m4.m[2][0] + v3.y * m4.m[2][1] + v3.z * m4.m[2][2]);
        }


        /*! \brief Transform a normalized plane by a transformation matrix.

            \param vPlane - xyz components are the plane normal, which this routine assumes is normalized.
                        w component is the distance of the plane from the origin, along
                        the direction of the normal.

            \param m4 - transformation matrix.

            \return Transformed plane as a 4-vector (same format as vPlane).

        */
        inline static Vec4 TransformPlane(const Vec4& vPlane, const Mat4& m4)
        {
            const Vec3& vPlaneNormalOrig = (Vec3&) vPlane;
            const float& fDistAlongDirOrig = vPlane.w;
            Vec3 vPlaneNormalXformed = TransformNormalByTranspose(vPlaneNormalOrig, m4);
            vPlaneNormalXformed.Normalize();
            const Vec4 vPoint32_tOnPlaneOrig = Vec4(- vPlaneNormalOrig * fDistAlongDirOrig, 1.0f);
            const Vec4 vPoint32_tOnPlaneXformed = vPoint32_tOnPlaneOrig * m4;
            const float fDistAlongDirXformed = vPlaneNormalXformed * (Vec3&) vPoint32_tOnPlaneXformed;
            return Vec4(vPlaneNormalXformed, -fDistAlongDirXformed);
        }


        inline bool operator == (const Mat4& rhs) const
        {
            return ((m[0][0] == rhs.m[0][0]) && (m[0][1] == rhs.m[0][1]) && (m[0][2] == rhs.m[0][2]) && (m[0][3] == rhs.m[0][3]) &&
                    (m[1][0] == rhs.m[1][0]) && (m[1][1] == rhs.m[1][1]) && (m[1][2] == rhs.m[1][2]) && (m[1][3] == rhs.m[1][3]) &&
                    (m[2][0] == rhs.m[2][0]) && (m[2][1] == rhs.m[2][1]) && (m[2][2] == rhs.m[2][2]) && (m[2][3] == rhs.m[2][3]) &&
                    (m[3][0] == rhs.m[3][0]) && (m[3][1] == rhs.m[3][1]) && (m[3][2] == rhs.m[3][2]) && (m[3][3] == rhs.m[3][3]));
        }

        inline bool operator != (const Mat4& rhs) const { return ! ((*this) == rhs); }

        inline Mat4 operator*(const Mat4& rhs) const    // Matrix-matrix multiplication
        {
            return Mat4(RowDotCol(0, rhs, 0), RowDotCol(0, rhs, 1), RowDotCol(0, rhs, 2), RowDotCol(0, rhs, 3),
                        RowDotCol(1, rhs, 0), RowDotCol(1, rhs, 1), RowDotCol(1, rhs, 2), RowDotCol(1, rhs, 3),
                        RowDotCol(2, rhs, 0), RowDotCol(2, rhs, 1), RowDotCol(2, rhs, 2), RowDotCol(2, rhs, 3),
                        RowDotCol(3, rhs, 0), RowDotCol(3, rhs, 1), RowDotCol(3, rhs, 2), RowDotCol(3, rhs, 3));
        }

        //inline Mat4  operator^(const Mat4 & rhs) const        { return ...; }  // maybe use this op for matrix outer product(?)

        inline Mat4 GetTranspose() const
        {
            return Mat4(m[0][0], m[1][0], m[2][0], m[3][0],
                        m[0][1], m[1][1], m[2][1], m[3][1],
                        m[0][2], m[1][2], m[2][2], m[3][2],
                        m[0][3], m[1][3], m[2][3], m[3][3]);
        }


        /*! \brief Assign scaling matrix
            \param scaleX - x component of scale
            \param scaleY - y component of scale
            \param scaleZ - z component of scale
        */
        inline void SetScale(const float& scaleX, const float& scaleY, const float& scaleZ)
        {
            m[0][0] = scaleX; m[0][1] =   0.0f; m[0][2] =   0.0f; m[0][3] = 0.0f;
            m[1][0] =   0.0f; m[1][1] = scaleY; m[1][2] =   0.0f; m[1][3] = 0.0f;
            m[2][0] =   0.0f; m[2][1] =   0.0f; m[2][2] = scaleZ; m[2][3] = 0.0f;
            m[3][0] =   0.0f; m[3][1] =   0.0f; m[3][2] =   0.0f; m[3][3] = 1.0f;
        }


        /*! \brief Assign scaling matrix
            \param scale - scale vector
        */
        inline void SetScale(const Vec3& vScale)
        {
            m[0][0] = vScale.x; m[0][1] =     0.0f; m[0][2] =     0.0f; m[0][3] = 0.0f;
            m[1][0] =     0.0f; m[1][1] = vScale.y; m[1][2] =     0.0f; m[1][3] = 0.0f;
            m[2][0] =     0.0f; m[2][1] =     0.0f; m[2][2] = vScale.z; m[2][3] = 0.0f;
            m[3][0] =     0.0f; m[3][1] =     0.0f; m[3][2] =     0.0f; m[3][3] = 1.0f;
        }


        /*! \brief Apply uniform scale to upper 3x3 part of matrix
            \param scale - scale factor
        */
        inline void ApplyScaleToUpper3x3(const float& scale)
        {
            m[0][0] *= scale; m[0][1] *= scale; m[0][2] *= scale;
            m[1][0] *= scale; m[1][1] *= scale; m[1][2] *= scale;
            m[2][0] *= scale; m[2][1] *= scale; m[2][2] *= scale;
        }


        /*! \brief Assign rightward unit vector part of transformation matrix
            \param vRight - vector point32_ting along transformed space +x axis
        */
        inline void SetRight(const Vec4& vRight)
        {
            m[0][0] = vRight.x; m[0][1] = vRight.y; m[0][2] = vRight.z; m[0][3] = vRight.w;
        }


        /*! \brief Assign forward unit vector part of transformation matrix
            \param vForward - vector point32_ting along transformed space +y axis
        */
        inline void SetForward(const Vec4& vForward)
        {
            m[1][0] = vForward.x; m[1][1] = vForward.y; m[1][2] = vForward.z; m[1][3] = vForward.w;
        }


        /*! \brief Assign upward unit vector part of transformation matrix
            \param vUp - vector point32_ting along transformed space +z axis
        */
        inline void SetUp(const Vec4& vUp)
        {
            m[2][0] = vUp.x; m[2][1] = vUp.y; m[2][2] = vUp.z; m[2][3] = vUp.w;
        }


        /*! \brief Assign translaton part of transformation matrix
            \param trans - translation vector
        */
        inline void SetTranslation(const Vec4& trans)
        {
            m[3][0] = trans.x; m[3][1] = trans.y; m[3][2] = trans.z; m[3][3] = trans.w;
        }


        /*! \brief Obtain "right" part of matrix
            \return Rightward vector
        */
        inline Vec4 GetRight(void) const
        {
            return Vec4(m[0][0], m[0][1], m[0][2], m[0][3]);
        }


        /*! \brief Obtain "forward" part of matrix
            \return Forwrd vector
        */
        inline Vec4 GetForward(void) const
        {
            return Vec4(m[1][0], m[1][1], m[1][2], m[1][3]);
        }


        /*! \brief Obtain "up" part of matrix
            \return Upward vector
        */
        inline Vec4 GetUp(void) const
        {
            return Vec4(m[2][0], m[2][1], m[2][2], m[2][3]);
        }


        /*! \brief Obtain translaton part of matrix
            \return Translation vector
        */
        inline Vec4 GetTranslation(void) const
        {
            return Vec4(m[3][0], m[3][1], m[3][2], m[3][3]);
        }


        /*! \brief Assign rotation matrix through angle about X axis
            \param angle - angle (in radians) to rotate
        */
        inline void SetRotationX(const float& angle)
        {
            float fSinVal = sin(angle);
            float fCosVal = cos(angle);

            m[0][0] = 1.0f;
            m[0][1] = 0.0f;
            m[0][2] = 0.0f;
            m[0][3] = 0.0f;
            m[1][0] = 0.0f;
            m[1][1] = fCosVal;
            m[1][2] = -fSinVal;
            m[1][3] = 0.0f;
            m[2][0] = 0.0f;
            m[2][1] = fSinVal;
            m[2][2] = fCosVal;
            m[2][3] = 0.0f;
            m[3][0] = 0.0f;
            m[3][1] = 0.0f;
            m[3][2] = 0.0f;
            m[3][3] = 1.0f;
        }


        /*! \brief Assign rotation matrix through angle about Y axis
            \param angle - angle (in radians) to rotate
        */
        inline void SetRotationY(const float& angle)
        {
            float fSinVal = sin(angle);
            float fCosVal = cos(angle);

            m[0][0] = fCosVal;
            m[0][1] = 0.0f;
            m[0][2] = fSinVal;
            m[0][3] = 0.0f;
            m[1][0] = 0.0f;
            m[1][1] = 1.0f;
            m[1][2] = 0.0f;
            m[1][3] = 0.0f;
            m[2][0] = -fSinVal;
            m[2][1] = 0.0f;
            m[2][2] = fCosVal;
            m[2][3] = 0.0f;
            m[3][0] = 0.0f;
            m[3][1] = 0.0f;
            m[3][2] = 0.0f;
            m[3][3] = 1.0f;
        }


        /*! \brief Assign rotation matrix through angle about Z axis
            \param angle - angle (in radians) to rotate
        */
        inline void SetRotationZ(const float& angle)
        {
            float fSinVal = sin(angle);
            float fCosVal = cos(angle);

            m[0][0] = fCosVal;
            m[0][1] = -fSinVal;
            m[0][2] = 0.0f;
            m[0][3] = 0.0f;
            m[1][0] = fSinVal;
            m[1][1] = fCosVal;
            m[1][2] = 0.0f;
            m[1][3] = 0.0f;
            m[2][0] = 0.0f;
            m[2][1] = 0.0f;
            m[2][2] = 1.0f;
            m[2][3] = 0.0f;
            m[3][0] = 0.0f;
            m[3][1] = 0.0f;
            m[3][2] = 0.0f;
            m[3][3] = 1.0f;
        }


        /*! \brief Assign rotation matrix through angle about X, Y and Z axes
            \param ax - angle (in radians) to rotate about X axis
            \param ay - angle (in radians) to rotate about Y axis
            \param az - angle (in radians) to rotate about Z axis
        */
        inline void SetRotationXYZ(const float& ax, const float& ay, const float& az)
        {
            float fSinX = sin(ax);
            float fCosX = cos(ax);
            float fSinY = sin(ay);
            float fCosY = cos(ay);
            float fSinZ = sin(az);
            float fCosZ = cos(az);

            m[0][0] = fCosY * fCosZ;
            m[0][1] = -fCosY * fSinZ;
            m[0][2] = fSinY;
            m[0][3] = 0.0f;
            m[1][0] = fSinX * fSinY * fCosZ + fCosX * fSinZ;
            m[1][1] = -fSinX * fSinY * fSinZ + fCosX * fCosZ;
            m[1][2] = -fSinX * fCosY;
            m[1][3] = 0.0f;
            m[2][0] = -fCosX * fSinY * fCosZ + fSinX * fSinZ;
            m[2][1] = fCosX * fSinY * fSinZ + fSinX * fCosZ;
            m[2][2] = fCosX * fCosY;
            m[2][3] = 0.0f;
            m[3][0] = 0.0f;
            m[3][1] = 0.0f;
            m[3][2] = 0.0f;
            m[3][3] = 1.0f;
        }


        /*! \brief Assign rotation matrix from axis and angle
            \param axis - axis about which to rotate
            \param angle - angle (in radians) to rotate
                http://www.mines.edu/~gmurray/ArbitraryAxisRotation/ArbitraryAxisRotation.html

            \note   Beware that the formulae given in the Matrix and Quaternion FAQ
                    for the diagonal terms appear to be nonsense.  Those formulae
                    result in non-unitary transforms.  (That FAQ often has errors.)

        */
        inline void SetRotation(const Vec3& axis, const float& angle)
        {
	        const float c   = cos(angle);
	        const float s   = sin(angle);
	        const float xs  = axis.x * s;
	        const float ys  = axis.y * s;
	        const float zs  = axis.z * s;
	        const float xyc = axis.x * axis.y * (1.0f - c);
	        const float xzc = axis.x * axis.z * (1.0f - c);
	        const float yzc = axis.y * axis.z * (1.0f - c);
	        m[0][0] =  axis.x * axis.x + (axis.y * axis.y + axis.z * axis.z) * c;
	        m[0][1] =  zs + xyc;
	        m[0][2] = -ys + xzc;

	        m[1][0] = -zs + xyc;
	        m[1][1] =    axis.y * axis.y + (axis.x * axis.x + axis.z * axis.z) * c;
	        m[1][2] =  xs + yzc;

	        m[2][0] =  ys + xzc;
	        m[2][1] = -xs + yzc;
	        m[2][2] =    axis.z * axis.z + (axis.x * axis.x + axis.y * axis.y) * c;

            m[0][3] =
            m[1][3] =
            m[2][3] =
            m[3][0] =
            m[3][1] =
            m[3][2] = 0.0f;

            m[3][3] = 1.0f;
        }

        inline void Orthonormalize()
        {
            Vec4 kU(m[0][0], m[0][1], m[0][2], m[0][3]);
            Vec4 kV(m[1][0], m[1][1], m[1][2], m[0][3]);
            Vec4 kW(m[2][0], m[2][1], m[0][2], m[2][3]);

            kU.Normalize();
            kW = kU ^ kV;
            kW.Normalize();
            kV = kW ^ kU;

            m[0][0] = kU[0];
            m[0][1] = kU[1];
            m[0][2] = kU[2];
            m[1][0] = kV[0];
            m[1][1] = kV[1];
            m[1][2] = kV[2];
            m[2][0] = kW[0];
            m[2][1] = kW[1];
            m[2][2] = kW[2];
        }


        /*! \brief Extract camera position from view matrix
        */
        static Vec4 CameraPositionFromViewMatrix(const Mat4& xView)
        {
            Vec4 vPos(-xView.GetTranslation());
            vPos = Mat4::TransformNormal((Vec3&) vPos, xView);
            vPos.w = 1.0f;
            return vPos;
        }

        float m[4][4];

    private:

        inline float RowDotCol(const int row, const Mat4& m4, const int& col) const
        {
            return m[row][0] * m4.m[0][col] +
                   m[row][1] * m4.m[1][col] +
                   m[row][2] * m4.m[2][col] +
                   m[row][3] * m4.m[3][col];
        }
};

// Public variables --------------------------------------------------------------

static const Mat4 Mat4_xIdentity(1.0f, 0.0f, 0.0f, 0.0f,
                                 0.0f, 1.0f, 0.0f, 0.0f,
                                 0.0f, 0.0f, 1.0f, 0.0f,
                                 0.0f, 0.0f, 0.0f, 1.0f);

#endif
