/*! \file vec3.h

    \brief 3-vector class

    \author Copyright 2005-2008 UCF/FIEA/MJG; All rights reserved.
*/
#ifndef VEC3_H
#define VEC3_H

#include <float.h>
#include <Math.h>
#include "macros.h"
#include "util.h"

// Macros --------------------------------------------------------------

 //! \brief Cast an array of floats to a Vec3
#define V3( v3 ) (*(Vec3*)v3)

#undef VEC3_INLINE
#if 1
#define VEC3_INLINE __forceinline
#elif 0
#define VEC3_INLINE inline
#else   // for profiling how much time is spent on Vec3 operations
#define VEC3_INLINE __declspec(noinline)
#endif

// Types --------------------------------------------------------------

/*! \brief Utility class for 3-vectors
*/
struct Vec3
{
    public:
        VEC3_INLINE Vec3() {}

        //! \brief construct vector from floats
        //! \param fx - x component
        //! \param fy - y component
        //! \param fz - z component
        VEC3_INLINE Vec3( float fx, float fy, float fz ) { x = fx ; y = fy ; z = fz ; }
        VEC3_INLINE Vec3( const float * fa ) { x = fa[0] ; y = fa[1] ; z = fa[2] ; }

        // assignment operators
        //! \brief add vector to self
        VEC3_INLINE Vec3 & operator += ( const Vec3 & rhs )          { x += rhs.x ; y += rhs.y ; z += rhs.z ; return * this ; }
        //! \brief subtract vector from self
        VEC3_INLINE Vec3 & operator -= ( const Vec3 & rhs )          { x -= rhs.x ; y -= rhs.y ; z -= rhs.z ; return * this ; }
        //! \brief multiply scalar to self
        VEC3_INLINE Vec3 & operator *= ( float f )                   { x *= f ; y *= f ; z *= f ; return * this ; }
        //! \brief divide self by scalar
        VEC3_INLINE Vec3 & operator /= ( float f )                   { (*this) *= ( 1.0f / f )  ; return * this ; }

        // unary operators
        //! \brief unary plus -- does nothing
        VEC3_INLINE Vec3 operator + () const                         { return * this ; }
        //! \brief unary minus -- return negative of this vector
        VEC3_INLINE Vec3 operator - () const                         { return Vec3( -x , -y , -z ) ; }

        // binary operators
        //! \brief add two vectors
        VEC3_INLINE Vec3 operator + ( const Vec3 & rhs ) const       { return Vec3( x + rhs.x , y + rhs.y , z + rhs.z ) ; }

        //! \brief subtract two vectors
        VEC3_INLINE Vec3 operator - ( const Vec3 & rhs ) const       { return Vec3( x - rhs.x , y - rhs.y , z - rhs.z ) ; }
        //Vec3 operator - ( const Vec3 & rhs ) const       ; // For CPU performance profiling. See util.cpp.

        //! \brief multiply vector by scalar
        VEC3_INLINE Vec3 operator * ( float f ) const                { return Vec3( x * f , y * f , z * f ) ; }
        //! \brief divide vector by scalar
        VEC3_INLINE Vec3 operator / ( float f ) const                { return (*this) * ( 1.0f / f ) ; }

        //! \brief multiply vector by scalar
        friend Vec3 operator * ( float f , const Vec3 & v)      { return Vec3( v.x * f , v.y * f , v.z * f ) ; }

        //! \brief compare two vectors for equal values
        VEC3_INLINE bool operator == ( const Vec3 & rhs ) const      { return ( ( x == rhs.x ) && ( y == rhs.y ) && ( z == rhs.z ) ) ; }
        //! \brief compare two vectors for equal values and negate
        VEC3_INLINE bool operator != ( const Vec3 & rhs ) const      { return ( ( x != rhs.x ) || ( y != rhs.y ) || ( z != rhs.z ) ) ; }

        //! \brief determine whether all components of another vector strictly exceed those of this
        VEC3_INLINE bool operator < ( const Vec3 & rhs ) const       { return ( ( x < rhs.x ) && ( y < rhs.y ) && ( z < rhs.z ) ) ; }
        //! \brief determine whether all components of this vector strictly exceed those of another
        VEC3_INLINE bool operator > ( const Vec3 & rhs ) const       { return ( ( x > rhs.x ) && ( y > rhs.y ) && ( z > rhs.z ) ) ; }

        //! \brief determine whether all components of another vector exceed or equal those of this
        VEC3_INLINE bool operator <= ( const Vec3 & rhs ) const
        {
            return ( ( x <= rhs.x ) && ( y <= rhs.y ) && ( z <= rhs.z ) ) ;
        }
        //! \brief determine whether all components of this vector exceed or equal those of another
        VEC3_INLINE bool operator >= ( const Vec3 & rhs ) const
        {
            return ( ( x >= rhs.x ) && ( y >= rhs.y ) && ( z >= rhs.z ) ) ;
        }

        //! \brief compute inner (i.e. dot) product of two vectors
        VEC3_INLINE float operator*( const Vec3 & rhs ) const        { return x * rhs.x + y * rhs.y + z * rhs.z ; }
        //! \brief compute outer (i.e. cross) product of two vectors
        VEC3_INLINE Vec3  operator^( const Vec3 & rhs ) const        { return Vec3( y * rhs.z - z * rhs.y , z * rhs.x - x * rhs.z , x * rhs.y - y * rhs.x ) ; }

        //! \brief return magnitude squared of a vector
        VEC3_INLINE float Mag2() const                               { return x * x + y * y + z * z ; }
        //! \brief return magnitude
        VEC3_INLINE float Magnitude() const                          { return sqrt( x * x + y * y + z * z ) ; }
        //! \brief normalize a vector, i.e. make its length unity
        VEC3_INLINE Vec3  GetDir() const                             { const float len = Magnitude() ; ASSERT( len > FLT_MIN ) ; return (*this) / len ; }
        //! \brief normalize a vector, i.e. make its length unity
        VEC3_INLINE void  Normalize()                                { * this = GetDir() ; }
        //! \brief Assign a vector to zero
        VEC3_INLINE void  Zero()                                     { x = y = z = 0.0f ; }

        VEC3_INLINE bool IsNormalized( void ) const
        {
            const float mag2 = Mag2() ;
            if( ( mag2 < 1.00001 ) && ( mag2 > 0.99999 ) ) return true ;
            return false ;
        }

        VEC3_INLINE bool Resembles( const Vec3 & other ) const
        {
            const Vec3          vDiff       = other - ( * this ) ;
            const float         fDiffMag2   = vDiff.Mag2() ;
            static const float  sDiffTol    = 1.0e-6f ;
            if( fDiffMag2 >= sDiffTol )
            {   // Absolute difference exceeds tolerance.
                const float fMag2 = this->Mag2() ;
                // Use relative difference.
                return ( fDiffMag2 / fMag2 < sDiffTol ) ;
            }
            return ( fDiffMag2 < sDiffTol ) ;
        }

        /*! \brief Return whether 3 points are colinear

            \param v0 - point in space
            \param v1 - point in space
            \param v2 - point in space

            \return Whether 3 given points are colinear (within tolerance)
        */
        static VEC3_INLINE bool IsColinear( const Vec3 & v0 , const Vec3 & v1 , const Vec3 & v2 )
        {
            const Vec3  v01     = v1 - v0 ;
            const Vec3  v02     = v2 - v0 ;
            const Vec3  cross   = v01 ^ v02 ;
            const float crosMag = cross.Mag2() ;    // area of parallelogram
            return crosMag < 1.0e-6 ;
        }


        /*! \brief Return whether 4 points are coplanar

            \param v0 - point in space
            \param v1 - point in space
            \param v2 - point in space
            \param v3 - point in space

            \return Whether 4 given points are coplanar (within tolerance)
        */
        static VEC3_INLINE bool IsCoplanar( const Vec3 & v0 , const Vec3 & v1 , const Vec3 & v2 , const Vec3 & v3 )
        {
            if( IsColinear( v0 , v1 , v2 ) )
            {   // If any 3 of the given points are colinear then these 4 points are colinear
                return true ;
            }
            const Vec3  v01     = v1 - v0 ;
            const Vec3  v02     = v2 - v0 ;
            const Vec3  v03     = v3 - v0 ;
            const Vec3  cross   = v01 ^ v02 ;
            const float triple  = FAbs( cross * v03 ) ; // volume of parallelepiped
            return triple < 1.0e-6 ;
        }

        float x ;   ///< x-component
        float y ;   ///< y-component
        float z ;   ///< z-component
} ;

// Public functions --------------------------------------------------------------

inline void Vec3Copy( float * vecDst , const float * vecSrc )
{
    vecDst[0] = vecSrc[0] ;
    vecDst[1] = vecSrc[1] ;
    vecDst[2] = vecSrc[2] ;
}


inline void Vec3Sum( float * vecSum , const float * vec1 , const float * vec2 )
{
    vecSum[0] = vec1[0] + vec2[0] ;
    vecSum[1] = vec1[1] + vec2[1] ;
    vecSum[2] = vec1[2] + vec2[2] ;
}


inline void Vec3Diff( float * vecDiff , const float * vec1 , const float * vec2 )
{
    vecDiff[0] = vec1[0] - vec2[0] ;
    vecDiff[1] = vec1[1] - vec2[1] ;
    vecDiff[2] = vec1[2] - vec2[2] ;
}


inline float Vec3DotProduct( const float * vec1 , const float * vec2 )
{
    const float v0 = vec1[0] * vec2[0] ;
    const float v1 = vec1[1] * vec2[1] ;
    const float v2 = vec1[2] * vec2[2] ;
    const float vSum = v0 + v1 + v2 ;
    return vSum ;
}


inline void Vec3CrossProduct( float * vecCross , const float * vec1 , const float * vec2 )
{
    vecCross[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1] ;
    vecCross[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2] ;
    vecCross[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0] ;
}


inline float Vec3Mag2( const float * vec )
{
    const float v0 = POW2( vec[0] ) ;
    const float v1 = POW2( vec[1] ) ;
    const float v2 = POW2( vec[2] ) ;
    const float vSum = v0 + v1 + v2 ;
    return vSum ; // not sure why, but this produces garbage: POW2( vec[0] ) + POW2( vec[1] ) + POW2( vec[3] ) ;
}


inline float Vec3Magnitude( const float * vec )
{
    return sqrt( Vec3Mag2( vec ) ) ;
}


inline bool Vec3IsNormalized( const float * vec )
{
    const float mag2 = Vec3Mag2( vec ) ;
    if( ( mag2 < 1.00001 ) && ( mag2 > 0.99999 ) ) return true ;
    return false ;
}


inline void Vec3Normalize( float * vecNormalized , const float * vec )
{
    const float magInverse = 1.0f / Vec3Magnitude( vec ) ;
    vecNormalized[0] = vec[0] * magInverse ;
    vecNormalized[1] = vec[1] * magInverse ;
    vecNormalized[2] = vec[2] * magInverse ;
    ASSERT( Vec3IsNormalized( vecNormalized ) ) ;
}


inline void Vec3Scale( float * vecScaled , const float * vec , float scale )
{
    vecScaled[0] = vec[0] * scale ;
    vecScaled[1] = vec[1] * scale ;
    vecScaled[2] = vec[2] * scale ;
}

#undef VEC3_INLINE

#endif