/*! \file vec4.h

    \brief 4-vector class

    \author Copyright 2005-2008 UCF/FIEA/MJG; All rights reserved.
*/
#ifndef VEC4_H
#define VEC4_H

#include <float.h>
#include "vec3.h"

// Macros --------------------------------------------------------------

//! \brief Cast an array of floats to a Vec4
#define V4( v4 ) (*(Vec4*)v4)

#undef VEC4_INLINE
#if 1
#define VEC4_INLINE __forceinline
#elif 0
#define VEC4_INLINE inline
#else   // for profiling how much time is spent on Vec4 operations
#define VEC4_INLINE __declspec(noinline)
#endif

// Types --------------------------------------------------------------

/*! \brief Utility class for 4-vectors
*/
struct Vec4
{
    public:
        VEC4_INLINE Vec4() {}

        //! \brief construct vector from floats
        //! \param fx - x component
        //! \param fy - y component
        //! \param fz - z component
        VEC4_INLINE Vec4( float fx, float fy, float fz )             { x = fx ; y = fy ; z = fz ; }
        //! \brief construct vector from floats
        //! \param fx - x component
        //! \param fy - y component
        //! \param fz - z component
        //! \param fw - w component
        VEC4_INLINE Vec4( float fx, float fy, float fz , float fw )  { x = fx ; y = fy ; z = fz ; w = fw ; }

        VEC4_INLINE Vec4( const Vec3 & v3 )                          { x = v3.x ; y = v3.y ; z = v3.z ; }
        VEC4_INLINE Vec4( const Vec3 & v3 , float fw )               { x = v3.x ; y = v3.y ; z = v3.z ; w = fw ; }

        VEC4_INLINE Vec4( const Vec4 & that)                         { x = that.x; y = that.y; z = that.z; w = that.w; }

        //! \brief construct vector from float array
        VEC4_INLINE Vec4( const float * pf )                         { x = pf[0] ; y = pf[1] ; z = pf[2] ; w = pf[3] ; }

        // assignment operators
        //! \brief add vector to self
        VEC4_INLINE Vec4 & operator += ( const Vec4 & rhs )          { x += rhs.x ; y += rhs.y ; z += rhs.z ; w += rhs.w ; return * this ; }
        VEC4_INLINE Vec4 & operator -= ( const Vec4 & rhs )          { x -= rhs.x ; y -= rhs.y ; z -= rhs.z ; w -= rhs.w ; return * this ; }
        VEC4_INLINE Vec4 & operator *= ( float f )                   { x *= f ; y *= f ; z *= f ; w *= f ; return * this ; }
        VEC4_INLINE Vec4 & operator /= ( float f )                   { (*this) *= ( 1.0f / f )  ; return * this ; }

        // unary operators
        VEC4_INLINE Vec4   operator + () const                       { return * this ; }
        VEC4_INLINE Vec4   operator - () const                       { return Vec4( -x , -y , -z , -w ) ; }

        // binary operators
        VEC4_INLINE Vec4   operator + ( const Vec4 & rhs ) const     { return Vec4( x + rhs.x , y + rhs.y , z + rhs.z , w + rhs.w ) ; }
        VEC4_INLINE Vec4   operator - ( const Vec4 & rhs ) const     { return Vec4( x - rhs.x , y - rhs.y , z - rhs.z , w - rhs.w ) ; }

        VEC4_INLINE Vec4   operator * ( float f ) const              { return Vec4( x * f , y * f , z * f , w * f ) ; }
        VEC4_INLINE Vec4   operator / ( float f ) const              { return (*this) * ( 1.0f / f ) ; }

        friend Vec4   operator * ( float f , const Vec4 & v)    { return Vec4( v.x * f , v.y * f , v.z * f , v.w * f ) ; }

        VEC4_INLINE bool   operator == ( const Vec4 & rhs ) const    { return ( ( x == rhs.x ) && ( y == rhs.y ) && ( z == rhs.z ) && ( w == rhs.w ) ) ; }
        VEC4_INLINE bool   operator != ( const Vec4 & rhs ) const    { return ( ( x != rhs.x ) || ( y != rhs.y ) || ( z != rhs.z ) || ( w != rhs.w ) ) ; }

        VEC4_INLINE float & operator[]( int i ) const                { return ((float*)(&x))[i] ; } ;

        //! \brief compute inner (i.e. dot) product of two vectors
        VEC4_INLINE float operator*( const Vec4 & rhs ) const        { return x * rhs.x + y * rhs.y + z * rhs.z + w * rhs.w ; }
        //! \brief compute outer (i.e. cross) product of two vectors, treating both as 3-vectors
        //! \note w-component is ignored and remains unchanged.
        VEC4_INLINE Vec4  operator^( const Vec4 & rhs ) const        { return Vec4( y * rhs.z - z * rhs.y , z * rhs.x - x * rhs.z , x * rhs.y - y * rhs.x , 0.0f ) ; }

        //! \brief return magnitude squared of a vector
        VEC4_INLINE float Mag2( void ) const                         { return x * x + y * y + z * z + w * w ; }
        //! \brief return magnitude squared of a vector, treating it as a 3-vector
        //! \note w-component is ignored
        VEC4_INLINE float Mag2v3( void ) const                       { return x * x + y * y + z * z ; }
        //! \brief return magnitude
        VEC4_INLINE float Magnitude( void ) const                    { return sqrt( Mag2() ) ; }
        //! \brief return magnitude of 3-vector
        //! \note w-component is ignored
        VEC4_INLINE float Magnitudev3( void ) const                  { return sqrt( Mag2v3() ) ; }
        //! \brief normalize a vector, i.e. make its length unity
        VEC4_INLINE void  Normalize( void )                          { const float len = Magnitude()   ; ASSERT( len > FLT_MIN ) ; (*this) /= len ; }
        //! \brief normalize first 3 components of a 4-vector, i.e. make its length unity
        //! \note w-component is ignored
        VEC4_INLINE void  Normalizev3( void )                        { const float len = Magnitudev3() ; ASSERT( len > FLT_MIN ) ; x /= len ; y /= len ; z /= len ; }
        VEC4_INLINE void  Zero( void )                               { x = y = z = w = 0.0f ; }
        //static VEC4_INLINE Vec4 ZERO( void )                         { return Vec4( 0.0f , 0.0f , 0.0f , 0.0f ) ; }
        VEC4_INLINE bool  IsNormalized( void ) const
        {
            const float mag = Mag2() ;
            return ( ( mag < 1.00001f ) && ( mag > 0.99999f ) ) ;
        }

        VEC4_INLINE bool  IsNormalizedv3( void ) const
        {
            const float mag = Mag2v3() ;
            return ( ( mag < 1.00001f ) && ( mag > 0.99999f ) ) ;
        }

        VEC4_INLINE bool Resembles( const Vec4 & other , const float fTolerance = 0.00001f ) const
        {
            Vec4 diff = other - ( * this ) ;
            const float diffMag2 = diff.Mag2() ;
            return ( diffMag2 < fTolerance ) ;
        }

        float DistFromPlane( const Vec3 & vPoint ) const
        {
            ASSERT( IsNormalizedv3() ) ;
            const float distFromPlane = vPoint * (Vec3&) *this + w ;
            return distFromPlane ;
        }

        float x ;   ///< x-component
        float y ;   ///< y-component
        float z ;   ///< z-component
        float w ;   ///< w-component
} ;

// Public variables --------------------------------------------------------------
// Public functions --------------------------------------------------------------

inline Vec4 RandomSpread( const Vec4 & vSpread )
{
    return  Vec4( RandomSpread( vSpread.x )
                , RandomSpread( vSpread.y )
                , RandomSpread( vSpread.z )
                , RandomSpread( vSpread.w ) ) ;
};

#undef VEC4_INLINE

#endif
