#include "Scope.h"

RTTI_DEFINITIONS(Scope)

Scope::Scope(unsigned int intialSize) : mParent(NULL) 
{
	mTable.Reserve(intialSize);
	mData.SetNumberOfBuckets(intialSize);
}
Scope::~Scope()
{
	Clear();

	// Orphan Yourself
	BecomeOrphan();
}

Scope::Scope(const Scope& source) : mParent(NULL)
{
	unsigned int newSize = source.Size();
	mTable.Reserve(newSize);
	mData.Rehash(newSize);
	this->operator=(source);
}

Scope& Scope::operator=(const Scope& source)
{
	Clear();
	mData.SetNumberOfBuckets(source.Size());
	CopyValues(source, false);
	return(*this);
}

bool Scope::operator==(const Scope& source)
{
	if(source.Size() != Size())
		return(false);
	for(unsigned int i = 0; i < mTable.Size(); i++){
		if(GetMemberName(i) != source.GetMemberName(i))
			return false;
		if(source.mTable[i]->second->GetType() == Datum::TYPE_SCOPE){ 		
			if(source[i].GetType() == Datum::TYPE_SCOPE){ // special case for scopes

				for(unsigned int j = 0; j < source[i].Size(); j++){
					if(*mTable[i]->second->Get<Datum::ScopeType>(j) != *source.mTable[i]->second->Get<Datum::ScopeType>(j))
						return false;
				}
			}
		}
		else
		{
			if(*mTable[i]->second != *source.mTable[i]->second)
				return false;
		}
	}
	return(true);
}
bool Scope::operator!=(const Scope& source)
{
	return(!operator==(source));
}
Datum& Scope::operator[](const string index)
{
	return(Append(index));
}
Datum& Scope::operator[](const unsigned int index) const
{
	assert(index < mTable.Size());
	return(*(mTable[index]->second));
}

const unsigned int Scope::Size() const
{
	return(mTable.Size());
}

// This should return the address of the Datum associated with the given name in this Scope
Datum* Scope::Find(const std::string name)
{
	Datum* foundDatum;
	foundDatum = NULL;
	
	HashMap<std::string, Datum*, HashString>::HashMapIterator myHashIterator = mData.Find(name);
	if(myHashIterator != mData.End()){
		return((*myHashIterator).second);
	}
	else
		return(NULL);
}


// This should return the address of the most-closely nested Datum associated with the given name in this Scope or its ancestors, if it exists, and 0 otherwise.  
Datum* Scope::Search(const std::string name, Scope** scopePointer)
{
	Datum* returnValue = Find(name);

	// Or ancestors	
	if(returnValue == NULL){
		if(mParent == NULL)
			return(NULL);
		else
			return mParent->Search(name, scopePointer);
	}
	else{
		if(scopePointer != NULL)
			(*scopePointer) = (this);
		return (returnValue);
	}
}

Datum& Scope::Append(const std::string name)
{
	Datum* foundDatum = Find(name);
	if(foundDatum == NULL){
		// Create an entry
		DictionaryEntry* newEntry = new DictionaryEntry();
		newEntry->first = name;
		
		// Create a Datum
		newEntry->second = new Datum();//datumReference;

		// Insert into the scope's look up table
		mData.Insert(*newEntry);
		
		// Insert into the scope table
		mTable.PushBack(newEntry);
		foundDatum = newEntry->second;
		newEntry = NULL;
	}
	return(*foundDatum);
}

Scope& Scope::AppendScope(const std::string name, const unsigned int index)
{
	Datum* foundDatum = Find(name);
	if(foundDatum){
		assert(foundDatum->GetType() == Datum::TYPE_SCOPE);
		assert(index <= foundDatum->Size()); // We don't want to create a new scope here since we could have dangling pointers
		if(index < foundDatum->Size()) // Return the scope that exists 
		{ 
			return(*foundDatum->Get<Datum::ScopeType>(index));
		}
		else //Special case where we create a new scope inside this scope, only gets here if size
		{
			Scope* newScope = new Scope();
			foundDatum->Set(newScope, index);
			return(*newScope);
		}
	}
	else{ // Need to create an outer scope

		//New Entry
		DictionaryEntry* newEntry = new DictionaryEntry();
		newEntry->first = name;
		
		// New Datum
		newEntry->second = new Datum();
		newEntry->second->SetType(Datum::TYPE_SCOPE);

		// New Scope
		Scope* newScope = new Scope();
		newEntry->second->Set(newScope);
		newScope->mParent = this;

		// Insert into the scope table
		mData.Insert(*newEntry);
		mTable.PushBack(newEntry);
		foundDatum = newEntry->second;
		newEntry = NULL;

		return(*newScope);
	}
}

void Scope::Adopt(Scope* childToAdopt, std::string nameOfChild, unsigned int indexToInsertAt)
{
	Scope* currentLocation = this;
	while(currentLocation->mParent != NULL){
		if(currentLocation == childToAdopt) 
			if((*currentLocation) == (*childToAdopt)) 
				assert(false); // would create a circular chain
		currentLocation = currentLocation->GetParent(); // next spot
	}
	childToAdopt->BecomeOrphan();
	childToAdopt->mParent = this;

	assert(Find(nameOfChild) == NULL);  // if there is already an entry in the scope with the passed name so assert.

	// Create a new entry, cannot use append and append scope since it won't be going at the end, will be going at a random spot in the table
	DictionaryEntry* newEntry = new DictionaryEntry();
	newEntry->first = nameOfChild;
	
	//Create a new datum for this entry
	newEntry->second = new Datum();

	// Place the stolen scope into the new datum
	newEntry->second->Set(childToAdopt);
	
	//Insert the new entry into this scope
	mData.Insert(*newEntry);
	mTable.InsertBefore(newEntry, indexToInsertAt);
}

Scope* Scope::GetParent() const
{
	return(mParent);
}

void Scope::CopyValues(const Scope& source, bool overrideConflicts)
{
	for(unsigned int i = 0; i < source.Size(); i++){ // Walk the right hand side's list and push them into the left hand side's hashmap
		// Delete all nested scopes
		if(source[i].GetType() == Datum::TYPE_SCOPE){
			for(unsigned int j = 0; j < source[i].Size(); j++){
				Scope* newScope = &AppendScope(source.GetMemberName(i), j); // create a scope
				
				newScope->CopyValues(*source[i].Get<Datum::ScopeType>(j));
				//*newScope = *source[i].Get<Datum::ScopeType>(j);
				newScope->mParent = this;
			}
		}
		else{
			Datum* newDatum = &Append(source.GetMemberName(i)); // create a datum
			*newDatum = source[i];
		}
	}
}

const unsigned int Scope::GetEntryIndex(std::string name)
{
	for(unsigned int i = 0; i < mTable.Size(); i++){
		if(mTable[i]->first == name)
			return i;
	}
	return(INT_MAX);
}

void Scope::Clear()
{
	mParent = NULL;
	//Deleting allocated data
	for(unsigned int i = 0; i < mTable.Size(); i++){
		if(mTable[i]->second->GetType() == Datum::TYPE_SCOPE){
			//Delete all nested scopes
			Datum * childScope = mTable[i]->second;
			for(unsigned int j = 0; j < childScope->Size(); j++){
				Scope * grandChildScope = childScope->Get<Datum::ScopeType>(j);
				delete grandChildScope; // Delete the created scope		
			}
		}
		delete mTable[i]->second; // Delete the created datum
		mTable[i]->second = NULL;
		delete mTable[i]; // Delete the created entry
		mTable[i] = NULL;
	}
	mData.Clear();
	mTable.Clear();
}

bool Scope::ReverseFind(const Scope* scopeReference, std::string& stringOutputReference)
{
	for(unsigned int i = 0; i < mTable.Size(); i++){
		if(mTable[i]->second->GetType() == Datum::TYPE_SCOPE){ 
			for(unsigned int j = 0; j < mTable[i]->second->Size(); j++){
				if((mTable[i]->second->Get<Datum::ScopeType>(j)) == (scopeReference)){	
					stringOutputReference = mTable[i]->first;
					return(true);
				}
			}
		}
	}
	return(false);
}

std::string Scope::GetMemberName(const unsigned int index) const
{
	assert(index < Size());
	return(mTable[index]->first);
}

void Scope::BecomeOrphan()
{
	if(mParent != NULL){
		std::string scopeName;
		mParent->ReverseFind(this,scopeName);
		if(scopeName.size() > 0){
			mParent->RemoveScopeEntry(scopeName, GetScopeIndex(this));
		}
		mParent = NULL;
	}
}

void Scope::RemoveScopeEntry(std::string name, const unsigned int index, bool killScope)
{
	if(index < Size()){
		Datum* datumReference = Find(name);
		if(Size() > 0){
			if(killScope){
				if(datumReference->GetType() == Datum::TYPE_SCOPE){
					delete datumReference->Get<Datum::ScopeType>(index);
				}
			}
			datumReference->RemoveIndex(index);
		}
	}
}

const unsigned int Scope::GetScopeIndex(Scope* scopeReference)
{
	if(mParent != NULL){
		for(unsigned int i = 0; i < mParent->mTable.Size(); i++){
			if(mParent->mTable[i]->second->GetType() == Datum::TYPE_SCOPE){ 
				for(unsigned int j = 0; j < mParent->mTable[i]->second->Size(); j++){
					if((mParent->mTable[i]->second->Get<Datum::ScopeType>(j)) == (scopeReference)){	
						return(j);
					}
				}
			}
		}
	}
	return(INT_MAX);
}

void Scope::RemoveEntry(std::string name)
{
	Datum* datumReference = Find(name);
	if(datumReference != NULL)
	{
		unsigned int deletedIndex = GetEntryIndex(name);

		delete mTable[deletedIndex]->second; // Delete the created datum
		mTable[deletedIndex]->second = NULL;
		
		delete mTable[deletedIndex]; // Delete the created entry
		mTable[deletedIndex] = NULL;


		mTable.Remove(deletedIndex);
		mData.Remove(name);
	}
}

Datum * Scope::GetEntry(unsigned int index)
{
	if(index >= Size())
		return NULL;
	return(mTable[index]->second);
}


IClonable * Scope::Clone()// const char * strCloneName , unsigned int index )
{
	// Use copy ctor to dupe this.
	Scope * pClone = new Scope( * this ) ;
	/*if( strCloneName )
	{   // New object is named.
		// Tell parent about clone.
		mParent->Adopt( pClone , strCloneName , index ) ;
	}*/
	return pClone ;
}

std::string Scope::GetOwnName()
{
	if(mParent){
		for(unsigned int i = 0; i < mParent->Size(); i++)
		{
			Datum * datumRef = mParent->GetEntry(i);
			if(datumRef->GetType() == Datum::TYPE_SCOPE)
			{
				for(unsigned int j = 0; j < datumRef->Size(); j++)
				{
					if((datumRef->Get<Datum::ScopeType>(j)) == this){
						return mParent->GetMemberName(i);
					}
				}
			}
		}
	}
	return std::string();
}