#ifndef __ENTITYTEST_H
#define __ENTITYTEST_H

#include <cxxtest/TestSuite.h>
#include "Vector.h"
#include "string.h"

#include "World.h"
#include "Entity.h"
#include "XmlParseMaster.h"
#include "XmlParseHelperEntity.h"

class PlayerEntityTest : public Entity
{
	RTTI_DECLARATIONS(PlayerEntityTest, Entity)


	PlayerEntityTest()
	{
		mySignatureInt.mName = ("My Int");
		mySignatureInt.mNumElements = 3;
		mySignatureInt.mType = Datum::TYPE_INT;
		mySignatureInt.mInit = new int(0);
		mSignatures.PushBack(mySignatureInt);

		mySignatureFloat.mName = ("My Float");
		mySignatureFloat.mNumElements = 3;
		mySignatureFloat.mType = Datum::TYPE_FLOAT;
		mySignatureFloat.mInit = new float(3.0f);
		mSignatures.PushBack(mySignatureFloat);

		arrayInt = new int[10];
		arrayInt[0] = 10;
		arrayInt[2] = 12;
		mySignatureExternal.mName = ("My External");
		mySignatureExternal.mNumElements = 10;
		mySignatureExternal.mType = Datum::TYPE_INT;
		mySignatureExternal.mStorage.TYPE_INT = arrayInt;
		mSignatures.PushBack(mySignatureExternal);

		//Scope				
		mySignatureScope.mName = ("My Scope");
		mySignatureScope.mNumElements = 2;
		mySignatureScope.mType = Datum::TYPE_SCOPE;
		mySignatureScope.mInit = new Scope();
		mSignatures.PushBack(mySignatureScope);

		
		mySignatureIntCopy = mySignatureInt;
		mySignatureIntCopy.mName = ("My Int Copy");
		mSignatures.PushBack(mySignatureIntCopy);

		mySignatureExternalCopy = mySignatureExternal;
		mySignatureExternalCopy.mName = ("My External Copy");
		mSignatures.PushBack(mySignatureExternalCopy);

		
		Populate();
	}

	virtual ~PlayerEntityTest()
	{
		delete mySignatureInt.mInit;
		delete mySignatureFloat.mInit;
		delete mySignatureScope.mInit;
		delete mySignatureExternal.mInit;
	}

	private:
		Signature mySignatureInt;
		Signature mySignatureFloat;
		Signature mySignatureExternal;

		//External
		Signature mySignatureIntCopy;
		Signature mySignatureExternalCopy;

		//Scope
		Signature mySignatureScope;

		//Junk
		int* arrayInt;
};
RTTI_DEFINITIONS(PlayerEntityTest)

EntityFactory( PlayerEntityTest )


/*!
*	A simple test suite: Just inherit CxxTest::TestSuite and write tests!
*/
class EntityUnitTest : public CxxTest::TestSuite
{
public:
	


	void setUp() {
		
	}
    void tearDown() {
		
	}


	/*!
	*	Tests the PopBack function
	*/
	void testEntity()
	{
		//Factories
		PlayerEntityTestFactory myPlayerFactory;

		//World
		World * myWorld = new World();;
		Sector * newSector = myWorld->AppendSector("Sector1");
		Sector * newSector2 = myWorld->AppendSector("Sector2");

		Scope * sectorList = myWorld->GetSectors();
		Scope & sectorList2 = myWorld->AppendScope("SectorList");

		Entity * playerE = newSector->NewEntity("PlayerEntityTest", "MyPlayer");

		TS_ASSERT_EQUALS(playerE->Is("Entity"), true);
		TS_ASSERT_EQUALS(playerE->Is("PlayerEntityTest"), true);

		TS_ASSERT_EQUALS(newSector2->Size(), 1);
		TS_ASSERT_EQUALS(sectorList->Size(), 2);
		TS_ASSERT_EQUALS(sectorList2.Size(), 2);

		myWorld->Process();

		delete myWorld;
	}
	
	void testEntityXMLParser()
	{
		//Factories
		PlayerEntityTestFactory myPlayerFactory;

		//New World
		World * myWorld = new World();

		//Shared Data
		XmlParseHelperScope::SharedDataScope* mySharedData = new XmlParseHelperScope::SharedDataScope(myWorld);

		//Master
		XmlParseMaster* myXMLParseMaster = new XmlParseMaster(mySharedData);

		//Helpers
		XmlParseHelperScope* myNewScopeHelper = new XmlParseHelperScope(myXMLParseMaster);
		XmlParseHelperEntity* myNewEntityHelper = new XmlParseHelperEntity(myXMLParseMaster);

		//Parse
		myXMLParseMaster->ParseFromFile("XML\\EntityTest.xml");

		//Depth Test
		TS_ASSERT_EQUALS(myXMLParseMaster->GetSharedData()->GetDepth(), 0);
		//Path Test
		TS_ASSERT_EQUALS(myXMLParseMaster->GetFileName(), "XML\\EntityTest.xml");
		
		//Content Test
		Scope & sectorList = myWorld->AppendScope("SectorList");
		Scope & sector1 = sectorList.AppendScope("sector1");
		Scope & entityList = sector1.AppendScope("EntityList");
		Scope & myPlayerEntity = entityList.AppendScope("MyPlayerEntity");

		PlayerEntityTest & myPlayer = (*myPlayerEntity.As<PlayerEntityTest>());

		//First Set
		Scope firstScope = myPlayerEntity.AppendScope("FirstScope"); 
		Datum firstInt = firstScope.Append("FirstInt");
		Datum firstVector = firstScope.Append("FirstVector");
		Datum firstMatrix = firstScope.Append("FirstMatrix");
		
		//Inner Set
		Scope innerScope = firstScope.AppendScope("InnerScope");
		Datum innerInt = innerScope.Append("InnerInt");
		Datum innerFloat = innerScope.Append("InnerFloat");
		Datum innerString = innerScope.Append("InnerString");

		TS_ASSERT_EQUALS(firstInt.Get<int>(), 1);
		TS_ASSERT_EQUALS(firstInt.Get<int>(1), 2);
		TS_ASSERT_EQUALS(firstInt.Get<int>(2), 3);


		TS_ASSERT_EQUALS(innerInt.Get<int>(), 1);
		TS_ASSERT_EQUALS(innerInt.Get<int>(1), 2);
		TS_ASSERT_EQUALS(innerFloat.Get<float>(), 1.5f);
		TS_ASSERT_EQUALS(innerString.Get<std::string>(), "Stuff");

		Vec4 myVector(3.9f, 1.5f, 2.1f, 1.0f);
		TS_ASSERT_EQUALS(myVector, firstVector.Get<Vec4>(0));

		Mat4 myMatrix(1.0f,2.0f,3.0f,4.0f,5.0f,6.0f,7.0f,8.0f,9.0f,10.0f,11.0f,12.0f,13.0f,14.0f,15.0f,16.0f);
		TS_ASSERT_EQUALS(myMatrix,  firstMatrix.Get<Mat4>());


		
		TS_ASSERT_EQUALS(myPlayer.IsPrescribedAttribute("My Int"), true);
		TS_ASSERT_EQUALS(myPlayer.IsAttribute("My Int"), true);
		TS_ASSERT_EQUALS(myPlayer.IsAuxiliaryAttribute("My Int"), false);
		Datum* intDatum = myPlayer.Find("My Int");
		TS_ASSERT_EQUALS(intDatum->Get<int>(0), 100);
		intDatum->Set(1,0);
		TS_ASSERT_EQUALS(intDatum->Get<int>(1), 0);
		TS_ASSERT_EQUALS(intDatum->Size(), 3);


		//External
		TS_ASSERT_EQUALS(myPlayer.IsAuxiliaryAttribute("My External"), false);
		Datum* externalDatum = myPlayer.Find("My External");

		TS_ASSERT_EQUALS(externalDatum->Get<int>(0), 10);
		TS_ASSERT_EQUALS(externalDatum->Get<int>(2), 12);

		
		//Copy
		TS_ASSERT_EQUALS(myPlayer.IsAuxiliaryAttribute("My Int Copy"), false);
		Datum* intDatumCopy = myPlayer.Find("My Int Copy");
		TS_ASSERT_EQUALS(intDatumCopy->Get<int>(0), 0);
		TS_ASSERT_EQUALS(intDatumCopy->Get<int>(1), 0);
		TS_ASSERT_EQUALS(intDatumCopy->Size(), 3);

		TS_ASSERT_EQUALS(myPlayer.IsAuxiliaryAttribute("My External"), false);
		Datum* externalDatumCopy = myPlayer.Find("My External");


		TS_ASSERT_EQUALS(externalDatumCopy->Get<int>(0), 10);
		TS_ASSERT_EQUALS(externalDatumCopy->Get<int>(2), 12);

		//Scope
		Datum* scopeDatum = myPlayer.Find("My Scope");
		scopeDatum->Get<Datum::ScopeType>(0)->Append("LOL");
		
		Datum* foundScope  = scopeDatum->Get<Datum::ScopeType>(0)->Find("LOL");
		foundScope->SetType(Datum::TYPE_INT);


		//Aux
		Datum newDatum = myPlayer.AppendAuxiliaryAttribute("New Attribute");
		newDatum.SetType(Datum::TYPE_INT);
		TS_ASSERT_EQUALS(8, myPlayer.AuxBegin());
		
		TS_ASSERT_EQUALS(myPlayer.IsPrescribedAttribute("New Attribute"), false);
		TS_ASSERT_EQUALS(myPlayer.IsAttribute("New Attribute"), true);
		TS_ASSERT_EQUALS(myPlayer.IsAuxiliaryAttribute("New Attribute"), true);

		myWorld->Process();

		delete myNewScopeHelper;
		delete myNewEntityHelper;
		delete mySharedData;
		delete myXMLParseMaster;
		delete myWorld;
	}
};

#endif // __ENTITYTEST_H